'''
Script to test use of KN1D, also integrating it within an Aurora impurity transport simulation.

It is recommended to run this in IPython.

sciortino, Jan 2021
'''

import numpy as np
import matplotlib.pyplot as plt
plt.ion()
import omfit_eqdsk, omfit_gapy
import sys, os
from scipy.interpolate import interp1d

# Make sure that package home is added to sys.path
sys.path.append('../')
import aurora

try: # pass any argument via the command line to show plots
    plot = len(sys.argv)>1
except:
    plot = False

# read in default Aurora namelist
namelist = aurora.default_nml.load_default_namelist()

# Use gfile and statefile in local directory:
examples_dir = os.path.dirname(os.path.abspath(__file__))
geqdsk = omfit_eqdsk.OMFITgeqdsk(examples_dir+'/example.gfile')
inputgacode = omfit_gapy.OMFITgacode(examples_dir+'/example.input.gacode')
# NB: this example.input.gacode file does not contain edge kinetic profiles and does not have a reasonable Te at the LCFS
# In this example, we will make up all kinetic data in the SOL based on exponential decay lengths.

# save kinetic profiles on a rhop (sqrt of norm. pol. flux) grid
kp = namelist['kin_profs']
kp['Te']['rhop'] = kp['ne']['rhop'] = np.sqrt(inputgacode['polflux']/inputgacode['polflux'][-1])
kp['ne']['vals'] = inputgacode['ne']*1e13 # 1e19 m^-3 --> cm^-3
kp['Te']['vals'] = inputgacode['Te']*1e3  # keV --> eV

# set impurity species and sources rate
imp = namelist['imp'] = 'C'
namelist['source_type'] = 'const'
namelist['Phi0'] = 2e20  # particles/s

# setup Aurora simulation without neutrals and CXR
asim = aurora.core.aurora_sim(namelist, geqdsk=geqdsk)

# estimate connection lengths from the EFIT g-EQDSK
clen_divertor_m, clen_limiter_m = aurora.estimate_clen(geqdsk)

# Estimate radial separation of boundary to separatrix and limiter to separatrix.
# This can be done by hand if desired, or using the following Aurora function for a specific device/shot/time
# combination using the EFIT a-EQDSK
#bound_sep_cm, lim_sep_cm = aurora.grids_utils.estimate_boundary_distance(my_shot, my_device, time_ms.)
bound_sep_cm = 1.5
lim_sep_cm = 1.0
p_H2_mTorr=0.05 # mTorr
innermost_rmid_cm=2.0

kn1d_res = aurora.run_kn1d(kp['ne']['rhop'], kp['ne']['vals'][0,:], kp['Te']['vals'][0,:], kp['Te']['vals'][0,:],
                           geqdsk, p_H2_mTorr, clen_divertor_m*1e2, clen_limiter_m*1e2,
                           bound_sep_cm, lim_sep_cm, innermost_rmid_cm, plot_kin_profs=True)

# series of plots to visualize (processed) KN1D output
aurora.kn1d.plot_overview(kn1d_res)     # overview of inputs and outputs
aurora.kn1d.plot_exc_states(kn1d_res)   # excited states
aurora.kn1d.plot_emiss(kn1d_res)        # Ly-a and D-a emission profiles
aurora.kn1d.plot_transport(kn1d_res)    # gradient scale lengths and "effective transport"

# add atomic neutral density to inputs and turn on charge exchange recombination
kp['n0']['rhop'] = kp['ne']['rhop']
n0 = interp1d(kn1d_res['kn1d_profs']['rhop'],kn1d_res['kn1d_profs']['n0'])(kp['Te']['rhop'])
kp['n0']['vals'] = n0[None,:] # need to set to (time,radius) grid, here time-independent
namelist['cxr_flag'] = True

# Now get aurora setup
asim_cxr = aurora.core.aurora_sim(namelist, geqdsk=geqdsk)

# set time-independent transport coefficients (flat D=1 m^2/s, V=-2 cm/s)
D_z = 1e4 * np.ones(len(asim.rvol_grid))  # cm^2/s
V_z = -2e2 * np.ones(len(asim.rvol_grid)) # cm/s

# run Aurora forward model with and without CXR
out = asim.run_aurora(D_z, V_z, plot=plot)
out_cxr = asim_cxr.run_aurora(D_z, V_z, plot=plot)

nz = out[0]
nz_cxr = out_cxr[0]

#ls_cycle = aurora.get_ls_cycle()
c_cycle = aurora.get_color_cycle()

fig, ax = plt.subplots()
for cs in np.arange(nz.shape[1]):
    col = next(c_cycle)
    ax.plot(asim.rhop_grid, nz[:,cs,-1], c=col, ls='--')
    ax.plot(asim_cxr.rhop_grid, nz_cxr[:,cs,-1], c=col, ls='-', label=f'{imp}{cs}+')
ax.plot([],[], 'w-', label=' ')
ax.plot([],[],'k-', label='with CXR')
ax.plot([],[],'k--', label='without CXR')
ax.legend(loc='best').set_draggable(True)  
ax.set_xlabel(r'$\rho_p$')
ax.set_ylabel(r'$n_z$ [$cm^{-3}$]')
ax.set_xlim([0.85,np.max(asim.rhop_grid)])


# add radiation
asim.rad = aurora.compute_rad(imp, nz.transpose(2,1,0), asim.ne, asim.Te,
                              prad_flag=True, thermal_cx_rad_flag=False)
asim_cxr.rad = aurora.compute_rad(imp, nz_cxr.transpose(2,1,0), asim.ne, asim.Te,
                                  n0=asim_cxr.n0,Ti=asim_cxr.Ti, 
                                  prad_flag=True, thermal_cx_rad_flag=True)

c_cycle = aurora.get_color_cycle()
fig, ax = plt.subplots()
for cs in np.arange(nz.shape[1]-1):  # fully-stripped ions have no line radiation
    col = next(c_cycle)
    ax.plot(asim.rhop_grid, asim.rad['line_rad'][-1,cs,:], c=col, ls='--')
    ax.plot(asim_cxr.rhop_grid, asim_cxr.rad['line_rad'][-1,cs,:], c=col, ls='-', label=f'{imp}{cs}+')
ax.plot([],[], 'w-', label=' ')
ax.plot([],[],'k-', label='with CXR')
ax.plot([],[],'k--', label='without CXR')
ax.legend(loc='best').set_draggable(True)  
ax.set_xlabel(r'$\rho_p$')
ax.set_ylabel(r'$P_{rad}$ [$W/m^3$]')
ax.set_xlim([0.85,np.max(asim.rhop_grid)])


# Check total radiated power components
rhop = np.linspace(np.min(asim.rhop_grid),1.0, 303)
ne = interp1d(asim.rhop_grid, asim.ne[0,:], kind='cubic')(rhop)
Te = interp1d(asim.rhop_grid, asim.Te[0,:], kind='cubic')(rhop)
Ti = interp1d(asim.rhop_grid, asim.Ti[0,:], kind='cubic')(rhop)
n0 = interp1d(asim.rhop_grid, asim_cxr.n0[0,:], kind='cubic')(rhop)

# get flux surface volumes and coordinates
grhop = np.sqrt(geqdsk['fluxSurfaces']['geo']['psin'])
gvol = geqdsk['fluxSurfaces']['geo']['vol']

# interpolate on our grid
vol = interp1d(grhop, gvol)(rhop)

nz_interp = np.zeros((len(rhop),nz.shape[1]))
nz_cxr_interp = np.zeros((len(rhop),nz.shape[1]))

for cs in np.arange(nz.shape[1]):
    nz_interp[:,cs] = interp1d(asim.rhop_grid, nz[:,cs,-1])(rhop)
    nz_cxr_interp[:,cs] = interp1d(asim.rhop_grid, nz_cxr[:,cs,-1])(rhop)

# get total radiated power components
rad_model = aurora.radiation_model(imp,rhop, ne, Te, vol, nz_cm3=nz_interp)
rad_model_cxr = aurora.radiation_model(imp, rhop, ne, Te, vol, nz_cm3=nz_cxr_interp,
                                       n0_cm3=n0, Ti_eV=Ti)

# print to screen 0-D prediction of ion total radiated power
print(f'Prad = {np.round(rad_model["Prad"]/1e6,3)} MW without neutrals')
print(f'Prad = {np.round(rad_model_cxr["Prad"]/1e6,3)} MW with neutrals')



import numpy as np,sys,os
from scipy.interpolate import interp1d, RectBivariateSpline
from . import grids_utils

def get_rhop_RZ(R,Z, geqdsk):
    '''Find rhop at every R,Z [m] based on the equilibrium in the geqdsk dictionary.
    '''
    return RectBivariateSpline(geqdsk['AuxQuantities']['Z'],
                               geqdsk['AuxQuantities']['R'],
                               geqdsk['AuxQuantities']['RHOpRZ']).ev(Z,R)



def vol_average(quant, rhop, method='omfit', geqdsk=None, device=None, shot=None, time=None, return_geqdsk=False):
    '''Calculate the volume average of the given radially-dependent quantity on a rhop grid. 

    Args:
        quant : array, (space, ...)
            quantity that one wishes to volume-average. The first dimension must correspond to space,
            but other dimensions may be exist afterwards.
        rhop : array, (space,)
            Radial rhop coordinate in cm units. 
        method : {'omfit','fs'}
            Method to evaluate the volume average. The two options correspond to the way to compute
            volume averages via the OMFIT fluxSurfaces classes and via a simpler cumulative sum in r_V 
            coordinates. The methods only slightly differ in their results. Note that 'omfit' will fail if 
            rhop extends beyond the LCFS, while method 'fs' can estimate volume averages also into the SOL.
            Default is method='omfit'. 
        geqdsk : output of the omfit_eqdsk.OMFITgeqdsk class, postprocessing the EFIT geqdsk file
            containing the magnetic geometry. If this is left to None, the function internally tries to fetch
            it using MDS+ and omfit_eqdsk. In this case, device, shot and time to fetch the equilibrium 
            are required. 
        device : str
            Device name. Note that routines for this device must be implemented in omfit_eqdsk for this 
            to work. 
        shot : int
             Shot number of the above device, e.g. 1101014019 for C-Mod.
        time : float
             Time at which equilibrium should be fetched in units of ms. 
        return_geqdsk : bool
             If True, omfit_eqdsk dictionary is also returned

    Returns:
        quant_vol_avg : array, (space, ...)
            Volume average of the quantity given as an input, in the same units as in the input.
            If extrapolation beyond the range available from EFIT volume averages over a shorter section
            of the radial grid will be attempted. This does not affect volume averages within the LCFS. 
        geqdsk : dict
            Only returned if return_geqdsk=True. 
    '''
    if time is not None and np.mean(time)<1e2:
        print('Input time was likely in the wrong units! It should be in [ms]!')
    if np.max(rhop)>1.0:
        print("Input rhop goes beyond the LCFS! Results may not be meaningful (and can only be obtained via method=='fs').")
        
    if geqdsk is None:
        # Fetch device geqdsk from MDS+ and post-process it using the OMFIT geqdsk format.
        try:
                import omfit_eqdsk
        except:
                raise ValueError('Could not import omfit_eqdsk! Install with pip install omfit_eqdsk')
        geqdsk = omfit_eqdsk.OMFITgeqdsk('').from_mdsplus(device=device, shot=shot,
                                                          time=time, SNAPfile='EFIT01',
                                                          fail_if_out_of_range=False,time_diff_warning_threshold=20)

    if method=='fs':
        # obtain mapping between rhop and r_V coordinates
        rho_pol, r_V_ = grids_utils.get_rhopol_rvol_mapping(geqdsk)
        
        # find r_V corresponding to input rhop (NB: extrapolation in the far SOL should be used carefully)
        r_V = interp1d(rho_pol, r_V_, bounds_error=False)(rhop)
        
        # use convenient volume averaging in r_V coordinates
        if np.any(np.isnan(r_V)):
            print('Ignoring all nan points! These may derive from an attempted extrapolation or from nan inputs')
            
        vol_avg = rV_vol_average(quant[~np.isnan(r_V)], r_V[~np.isnan(r_V)])
        
    elif method=='omfit':
        # use fluxSurfaces classes from OMFIT
        rhopp = np.sqrt(geqdsk['fluxSurfaces']['geo']['psin'])
        quantp = interp1d(rhop, quant, bounds_error=False,fill_value='extrapolate')(rhopp)
        vol_avg = geqdsk['fluxSurfaces'].volume_integral(quantp)

    else:
        raise ValueError('Input method for volume average could not be recognized')

    if return_geqdsk:
        return vol_avg, geqdsk
    else:
        return vol_avg


def rV_vol_average(quant,r_V):
    '''Calculate a volume average of the given radially-dependent quantity on a r_V grid.
    This function makes useof the fact that the r_V radial coordinate, defined as 
    :math:`r_V = \sqrt{ V / (2 \pi^2 R_{axis} }`,
    maps shaped volumes onto a circular geometry, making volume averaging a trivial 
    operation via
    :math:`\langle Q \rangle = \Sigma_i Q(r_i) 2 \pi \ \Delta r_V`
    where :math:`\Delta r_V` is the spacing between radial points in r_V.
    
    Note that if the input r_V coordinate is extended outside the LCFS,
    this function will return the effective volume average also in the SOL, since it is 
    agnostic to the presence of the LCFS. 

    Args:
        quant : array, (space, ...)
            quantity that one wishes to volume-average. The first dimension must correspond to r_V,
            but other dimensions may be exist afterwards. 
        r_V : array, (space,)
            Radial r_V coordinate in cm units. 

    Returns:
        quant_vol_avg : array, (space, ...)
            Volume average of the quantity given as an input, in the same units as in the input
    '''
    quant_vol_avg = 2.*np.cumsum(quant*r_V*np.diff(r_V,prepend=0.0)) / (r_V[-1]**2 )

    return quant_vol_avg


    
def rad_coord_transform(x,name_in,name_out, geqdsk):
    """Transform from one radial coordinate to another. Note that this coordinate conversion is only
    strictly valid inside of the LCFS.

    Args:
        x: array
            input x coordinate
        name_in: str
            input x coordinate name ('rhon','r_V','rhop','rhov','Rmid','rmid','r/a')
        name_out: str
            input x coordinate ('rhon','psin','r_V', 'rhop','rhov','Rmid','rmid','r/a')
        geqdsk: dict
            gEQDSK dictionary, as obtained from the omfit-eqdsk package. 
    
    Returns:
        Conversion of `x` for the requested radial grid coordinate.
    """
    if name_in == name_out:
        return x

    if 'r_V' not in geqdsk['fluxSurfaces']['geo']:
        R0 = geqdsk['RMAXIS']
        eq_vol = geqdsk['fluxSurfaces']['geo']['vol']
        r_V = np.sqrt(eq_vol/(2*np.pi**2*R0))
        geqdsk['fluxSurfaces']['geo']['r_V'] = r_V

    # sqrt(norm. tor. flux)
    rhon_ref = geqdsk['fluxSurfaces']['geo']['rhon']
    # norm. pol. flux
    psin_ref = geqdsk['fluxSurfaces']['geo']['psin']
    # sqrt(norm. pol. flux)
    rhop_ref = np.sqrt(psin_ref)
    #volume radius
    r_V = geqdsk['fluxSurfaces']['geo']['r_V']
    # R at midplane
    Rmid = geqdsk['fluxSurfaces']['midplane']['R']
    # r at midplane
    R0 = geqdsk['fluxSurfaces']['R0']
    rmid = Rmid - R0

    # Interpolate to transform coordiantes
    if name_in == 'rhon':
        coord_in = rhon_ref
    elif name_in == 'rhop':
        coord_in = rhop_ref
    elif name_in == 'r_V':
        coord_in = r_V
    elif name_in == 'rhov':
        r_V_lcfs = np.interp(1, rhon_ref, r_V)
        coord_in = r_V/r_V_lcfs
    elif name_in == 'Rmid':
        coord_in = Rmid
    elif name_in == 'rmid':
        coord_in = rmid
    elif name_in == 'r/a':
        rmid_lcfs = np.interp(1, rhon_ref, rmid)
        coord_in = rmid/rmid_lcfs
    else:
        raise ValueError('Input coordinate was not recognized!')

    if name_out == 'rhon':
        coord_out = rhon_ref
    elif name_out == 'psin':
        coord_out = psin_ref
    elif name_out == 'rhop':
        coord_out = rhop_ref
    elif name_out == 'r_V':
        coord_out = r_V
    elif name_out == 'rhov':
        r_V_lcfs = np.interp(1, rhon_ref, r_V)
        coord_out = r_V/r_V_lcfs
    elif name_out == 'Rmid':
        coord_out = Rmid
    elif name_out == 'rmid':
        coord_out = rmid
    elif name_out == 'r/a':
        rmid_lcfs = np.interp(1, rhon_ref, rmid)
        coord_out = rmid/rmid_lcfs
    else:
        raise ValueError('Output coordinate was not recognized!')

    ind = coord_in != 0

    #trick for better extrapolation
    return np.interp(x,coord_in[ind],coord_out[ind]/coord_in[ind])*x





name='aurora'
__version__="0.1.9"

import numpy as np, sys

# don't try to import compiled Fortran if building documentation or package:
if not np.any([('sphinx' in k and not 'sphinxcontrib' in k) for k in sys.modules]) and\
   not np.any([('distutils' in k.split('.') and 'command' in k.split('.')) for k in sys.modules]):
    from ._aurora import run,time_steps

from .core import *
from .atomic import *
from .adas_files import *

from .source_utils import *
from .default_nml import *
from .grids_utils import *
from .coords import *
from .radiation import *

from .particle_conserv import *
from .plot_tools import *
from .animate import *

from .janev_smith_rates import *
from .nbi_neutrals import *
from .neutrals import *

from .synth_diags import *

from .solps import *
from .kn1d import *



# -*-Python-*-
# Created by sciortinof at 31 Jul 2020  09:11

'''
Methods for neutral beam analysis, particularly in relation to impurity transport studies.
These script collects functions that should be device-agnostic.
'''
import numpy as np
import matplotlib.pyplot as plt
plt.ion()
from scipy.interpolate import RectBivariateSpline
import copy, itertools

from .janev_smith_rates import js_sigma


def get_neutrals_fsa(neutrals, geqdsk, debug_plots=True):
    """Compute charge exchange recombination for a given impurity with neutral beam components,
    obtaining rates in [:math:`s^-1`] units. This method expects all neutral components to be given in a
    dictionary with a structure that is independent of NBI model (i.e. coming from FIDASIM, NUBEAM, 
    pencil calculations, etc.).

    Args:
        neutrals : dict
            Dictionary containing fields
            {"beams","names","R","Z", beam1, beam2, etc.}
            Here beam1,beam2,etc. are the names in neutrals["beams"]. "names" are the names of each 
            beam component, e.g. 'fdens','hdens','halo', etc., ordered according to "names". 
            "R","Z" are the major radius and vertical coordinates [cm] on which neutral density components are 
            given in elements such as
    
            .. code-block:: python
            
                neutrals[beams[0]]["n=0"][name_idx]

            It is currently assumed that n=0,1 and 2 beam components are provided by the user. 

        geqdsk : gEQDSK post-processed dictionary, as given by the omfit_eqdsk package. 

        debug_plots : bool, optional
            If True, various plots are displayed. 

    Returns:
        neut_fsa : dict
             Dictionary of flux-surface-averaged (FSA) neutral densities, in the same units as in the input. 
             Similarly to the input "neutrals", this dictionary has a structure like

             .. code-block:: python
             
                 neutrals_ext[beam][f'n={n_level}'][name_idx]

    """

    beams = neutrals['beams']
    names = neutrals['names']
    zz = neutrals['Z'] / 1e2  # cm --> m
    rr = neutrals['R'] / 1e2  # cm --> m

    if debug_plots:
        # for debugging/plotting:
        RBBBS = geqdsk['RBBBS']
        ZBBBS = geqdsk['ZBBBS']
        fig, ax = plt.subplots()
        ax.contourf(rr, zz, neutrals[beams[0]]['n=0'][0].T)
        ax.plot(RBBBS, ZBBBS, 'w--', lw=5)
        ax.set_xlabel('R [m]')
        ax.set_ylabel('Z [m]')

    # simple way to find rhop at any combination of R,Z coords:
    RHOpRZ = geqdsk['AuxQuantities']['RHOpRZ']
    Rgrid = geqdsk['AuxQuantities']['R']  # m
    Zgrid = geqdsk['AuxQuantities']['Z']  # m

    # extrapolate beam to cover the entire 2D poloidal cross section,
    # setting all neutral populations to 0 outside of the simulated region
    neutrals_ext = {}
    for beam in beams:
        neutrals_ext[beam] = {}
        for n_level in [0, 1, 2]:
            neutrals_ext[beam][f'n={n_level}'] = {}
            for ii, name in enumerate(names):
                dens = neutrals[beam][f'n={n_level}'][ii]

                f = interp2d(rr, zz, dens.T, kind='linear', bounds_error=False, fill_value=0.0)
                tmp = f(Rgrid, Zgrid)

                tmp[tmp < 0] = 0.0
                neutrals_ext[beam][f'n={n_level}'][name] = tmp

    if debug_plots:
        fig, ax = plt.subplots()
        CS = ax.contourf(Rgrid, Zgrid, neutrals_ext[beam]['n=0']['fdens'])
        cbar = fig.colorbar(CS)
        cbar.ax.set_ylabel(r'f density [$cm^{-3}$]')
        ax.plot(RBBBS, ZBBBS, c='w', lw=5)
        ax.axis('equal')
        CS = plt.contour(Rgrid, Zgrid, RHOpRZ, np.linspace(0.0, 1.2, 13), c='w')
        plt.clabel(CS, inline=1, fontsize=14)
        ax.set_xlabel('R [m]')
        ax.set_ylabel('Z [m]')

    # Get flux surface average quantities:
    neut_fsa = {}
    rhop = neut_fsa['rhop'] = geqdsk['AuxQuantities']['RHOp']
    neut_fsa['names'] = names

    for beam in beams:
        neut_fsa[beam] = {}

        for n_level in [0, 1, 2]:
            neut_fsa[beam][f'n={n_level}'] = {}

            for ii, name in enumerate(names):
                dens = neutrals_ext[beam][f'n={n_level}'][name]

                # flux surface average
                def avg_function(r, z):
                    return RectBivariateSpline(Zgrid, Rgrid, dens, kx=1, ky=1).ev(z, r)

                neut_fsa[beam][f'n={n_level}'][name] = geqdsk['fluxSurfaces'].surfAvg(function=avg_function)

    for beam in beams:
        # store beam component energies
        neut_fsa[beam]['f_energy'] = neutrals[beam]['f_energy']  # keV
        neut_fsa[beam]['h_energy'] = neutrals[beam]['h_energy']  # keV
        neut_fsa[beam]['t_energy'] = neutrals[beam]['t_energy']  # keV
        neut_fsa[beam]['m_beam'] = neutrals[beam]['m_beam']  # amu
    neut_fsa['m_bckg'] = copy.deepcopy(neutrals['m_bckg'])  # amu

    if debug_plots:
        ls_cycle = get_ls_cycle()

        # plot FSA neutral densities
        fig = plt.figure()
        fig.set_size_inches(12, 9, forward=True)
        a1 = plt.subplot2grid((4, 4), (0, 0), rowspan=4, colspan=3)
        a2 = plt.subplot2grid((4, 4), (0, 3), rowspan=4, colspan=1)
        for beam in beams:
            for n_level in [0, 1, 2]:
                for name in ['fdens', 'hdens', 'tdens', 'dcx+halo']:
                    if name == 'dcx+halo':
                        name = 'halo'  # simple renaming
                    lss = next(ls_cycle)
                    a1.plot(rhop, neut_fsa[beam][f'n={n_level}'][name], lss, lw=2.0)
                    a2.plot([], [], lss, lw=2.0, label=f'{beam} {name}, n={n_level+1}')

        a1.set_xlabel(r'$\rho_{p}$')
        a1.set_ylabel('FSA density')
        a2.axis('off')
        a2.legend(fontsize=14)
        # fig.tight_layout()

    return neut_fsa


def get_ls_cycle():
    # create useful list of plotting styles
    color_vals = ['b', 'g', 'r', 'c', 'm', 'y', 'k']
    style_vals = ['-', '--', '-.', ':']
    ls_vals = []
    for s in style_vals:
        for c in color_vals:
            ls_vals.append(c + s)
            ls_cycle = itertools.cycle(ls_vals)
    return ls_cycle


def get_NBI_imp_cxr_q(neut_fsa, q, rhop_Ti, times_Ti, Ti_prof, include_fast=True, include_halo=True, debug_plots=False):
    """Compute flux-surface-averaged (FSA) charge exchange recombination for a given impurity with
    neutral beam components, applying appropriate Maxwellian averaging of cross sections and
    obtaining rates in [:math:`s^-1`] units. This method expects all neutral components to be given in a
    dictionary with a structure that is independent of NBI model.

    Note that while Ti may be time-dependent, with a time base given by times_Ti, the FSA
    neutrals are expected to be time-independent. Hence, the resulting CXR rates will only have
    time dependence that reflects changes in Ti, but not the NBI.

    Args:
        neut_fsa : dict
             Dictionary containing FSA neutral densities in the form that is output by :py:meth:`get_neutrals_fsa`.
        q : int or float
             Charge of impurity species
        rhop_Ti : array-like
             Sqrt of poloidal flux radial coordinate for Ti profiles.
        times_Ti : array-like
             Time base on which Ti_prof is given [s]. 
        Ti_prof : array-like
             Ion temperature profile on the rhop_Ti, times_Ti bases.
        include_fast : bool, optional
             If True, include CXR rates from fast NBI neutrals. Default is True. 
        include_halo : bool, optional
             If True, include CXR rates from themral NBI halo neutrals. Default is True. 
        debug_plots : bool, optional
             If True, plot several plots to assess the quality of the calculation. 

    Returns:
        rates : dict
            Dictionary containing CXR rates from NBI neutrals. This dictionary has analogous form to the 
            :py:meth:`get_neutrals_fsa` function, e.g. we have 

            .. code-block:: python

                rates[beam][f'n={n_level}']['halo']

            Rates are on a radial grid corresponding to the input neut_fsa['rhop']. 

    For details on inputs and outputs, it is recommendeded to look at the internal plotting functions. 

    """
    m_bckg = neut_fsa['m_bckg']  # amu
    rhop = neut_fsa['rhop']

    if len(times_Ti) == 1:
        Ti = np.atleast_2d(interp1d(rhop_Ti, Ti_prof[:, 0] / 1e3, bounds_error=False, fill_value=3e-3)(rhop)).T  # keV
    else:
        Ti = RectBivariateSpline(times_Ti, rhop_Ti, Ti_prof / 1e3)(times_Ti, rhop)  # keV

    # collect rates for each energy component and excited state (ONLY fast neutrals here)
    rates = {}
    rates['rhop'] = rhop
    rates['times'] = times_Ti
    rates['cxr_total'] = np.zeros((len(rhop), len(times_Ti)))

    # setup dictionaries here in case include_fast=False
    for beam in beams:
        rates[beam] = {}
        rates[beam]['cxr_total'] = np.zeros((len(rhop), len(times_Ti)))
        for n_level in [0, 1, 2]:
            rates[beam][f'n={n_level}'] = {}

    if include_fast:
        # compute impurity recombination rate with fast neutral populations
        for beam in beams:
            m_beam = neut_fsa[beam]['m_beam']  # amu
            for n_level in [0, 1, 2]:
                for comp in ['f', 'h', 't']:
                    rates[beam][f'n={n_level}'][comp] = {}

                    # energy of each beam component
                    energy = neut_fsa[beam][f'{comp}_energy']  # keV

                    # create cross section function only as a function of energy/amu for Maxwellian average
                    sigma_fun = lambda E_per_amu: js_sigma(E_per_amu, q, n1=n_level + 1, type='cx')  # cm^2

                    rate = bt_rate_maxwell_average(sigma_fun, Ti, energy, m_bckg, m_beam, n_level + 1.0)  # .T
                    rates[beam][f'n={n_level}'][comp] = rate * neut_fsa[beam][f'n={n_level}'][f'{comp}dens'][:, np.newaxis]

                    # we are eventually interested in the total:
                    rates[beam]['cxr_total'] += rates[beam][f'n={n_level}'][comp]

            # sum over all beams:
            rates['cxr_total'] += rates[beam]['cxr_total']

    if include_halo:
        # Now add recombination from halo neutrals (all thermal interactions) - n-unresolved for impurities
        for beam in beams:
            m_beam = neut_fsa[beam]['m_beam']  # amu

            for n_level in [0, 1, 2]:
                # use Janev & Smith rates for halos too...
                sigma_fun = lambda E_per_amu: js_sigma(E_per_amu, q, n1=n_level + 1.0, type='cx')  # cm^2

                # Maxwell average using a "beam" at thermal energy (WRONG)
                # rate1 = bt_rate_maxwell_average(sigma_fun, Ti, Ti, m_bckg, m_beam, n_level + 1.0)

                # use proper thermal-thermal averaging
                # the following function at the moment can only take time-independent Ti!
                rate = tt_rate_maxwell_average(sigma_fun, Ti[:, 0], m_bckg, m_beam, n_level + 1.0)

                # plt.figure()
                # plt.plot(rhop, rate1, label='bt')
                # plt.plot(rhop, rate, label='tt')
                # plt.xlabel(r'$\rho_p$')
                # plt.legend()

                rates[beam][f'n={n_level}']['halo'] = rate[:, None] * neut_fsa[beam][f'n={n_level}']['dcx+halo'][:, None]
                rates[beam]['cxr_total'] += rates[beam][f'n={n_level}']['halo']
                rates['cxr_total'] += rates[beam][f'n={n_level}']['halo']

    if debug_plots:
        ls_cycle = get_ls_cycle()

        fig = plt.figure()
        fig.set_size_inches(12, 9, forward=True)
        a1 = plt.subplot2grid((4, 4), (0, 0), rowspan=4, colspan=3)
        a2 = plt.subplot2grid((4, 4), (0, 3), rowspan=4, colspan=1)

        for beam in beams:
            for n_level in [0, 1, 2]:
                if include_fast:
                    for comp in ['f', 'h', 't']:
                        # time average to reduce number of lines
                        lss = next(ls_cycle)
                        a1.plot(rhop, np.mean(rates[beam][f'n={n_level}'][comp], axis=-1), lss)
                        a2.plot([], [], lss, lw=2.0, label=f'{beam} {comp}-energy, n={n_level+1}')
                if include_halo:
                    lss = next(ls_cycle)
                    a1.plot(rhop, np.mean(rates[beam][f'n={n_level}']['halo'], axis=-1), lss)
                    a2.plot([], [], lss, lw=2.0, label=f'{beam} halo, n={n_level+1}')

            a1.plot(rhop, np.mean(rates[beam]['cxr_total'], axis=-1), label=f'beam {beam} CXR total')
            a2.plot([], [], label=f'beam {beam} CXR total')

        a1.plot(rhop, rates['cxr_total'])
        a2.plot([], [], label=f'CXR total')
        a1.set_xlabel(r'$\rho_p$')
        a1.set_ylabel(fr'CXR rate (q={q}) [$s^{{-1}}$]')
        a2.legend(fontsize=14)
        a2.axis('off')
        fig.tight_layout()

    return rates


def beam_grid(uvw_src, axis, max_radius=255.0):
    """Method to obtain the 3D orientation of a beam with respect to the device.
    The uvw_src and (normalized) axis arrays may be obtained from the d3d_beams method
    of fidasim_lib.py in the FIDASIM module in OMFIT. 

    This is inspired by `beam_grid` in fidasim_lib.py of the FIDASIM module (S. Haskey) 
    in OMFIT. 
    """

    pos = uvw_src + 100 * axis
    rsrc = np.sqrt(uvw_src[0] ** 2 + uvw_src[1] ** 2)
    if rsrc < max_radius:
        print("Source radius:{} cannot be less then max_radius:{}".format(rsrc, max_radius))
        raise ValueError()
    dis = np.sqrt(np.sum((uvw_src - pos) ** 2.0))
    beta = np.arcsin((uvw_src[2] - pos[2]) / dis)
    alpha = np.arctan2((pos[1] - uvw_src[1]), (pos[0] - uvw_src[0]))
    gamma = 0.0

    # Find where the origin has to be along the beam injection
    # axis so that x=0 is at a radius of max_radius
    a = axis[0] ** 2 + axis[1] ** 2
    b = 2 * (uvw_src[0] * axis[0] + uvw_src[1] * axis[1])
    c = uvw_src[0] ** 2 + uvw_src[1] ** 2 - max_radius ** 2
    t = (-b - sqrt(b ** 2 - 4 * a * c)) / (2 * a)
    origin = uvw_src + t * axis
    return alpha, beta, gamma, origin


def rotation_matrix(alpha, beta, gamma):
    """See the table of all rotation possiblities, on the Tait Bryan side
    https://en.wikipedia.org/wiki/Euler_angles#Tait.E2.80.93Bryan_angles
    """
    a = alpha
    b = beta
    g = gamma
    sa = np.sin(a)
    ca = np.cos(a)
    sb = np.sin(b)
    cb = np.cos(b)
    sg = np.sin(g)
    cg = np.cos(g)
    R = np.zeros((3, 3), dtype=float)

    R[0, 0] = ca * cb
    R[0, 1] = ca * sb * sg - cg * sa
    R[0, 2] = sa * sg + ca * cg * sb
    R[1, 0] = cb * sa
    R[1, 1] = ca * cg + sa * sb * sg
    R[1, 2] = cg * sa * sb - ca * sg
    R[2, 0] = -sb
    R[2, 1] = cb * sg
    R[2, 2] = cb * cg
    
    return R


def uvw_xyz(u, v, w, origin, R):
    """
    Computes array elements by multiplying the rows of the first
    array by the columns of the second array. The second array
    must have the same number of rows as the first array has
    columns. The resulting array has the same number of rows as
    the first array and the same number of columns as the second
    array.
    
    See uvw_to_xyz in fidasim.f90
    """
    u, v, w = np.atleast_1d(u), np.atleast_1d(v), np.atleast_1d(w)
    orig_shape = u.shape
    order = 'C'
    uvw = np.transpose(np.array([u.flatten(order=order), v.flatten(order=order), w.flatten(order=order)]))
    uvw_shifted = np.transpose(uvw - origin[np.newaxis, :])
    basis = np.linalg.inv(R)

    xyz = np.dot(basis, uvw_shifted)
    x, y, z = xyz[0, :].reshape(orig_shape, order='C'), xyz[1, :].reshape(orig_shape, order='C'), xyz[2, :].reshape(orig_shape, order='C')
    return x, y, z


def xyz_uvw(x, y, z, origin, R):
    """
    Computes array elements by multiplying the rows of the first
    array by the columns of the second array. The second array
    must have the same number of rows as the first array has
    columns. The resulting array has the same number of rows as
    the first array and the same number of columns as the second
    array.
    
    See xyz_to_uvw in fidasim.f90
    """
    x, y, z = np.atleast_1d(x), np.atleast_1d(y), np.atleast_1d(z)
    orig_shape = x.shape
    order = 'C'
    xyz = np.array([x.flatten(order=order), y.flatten(order=order), z.flatten(order=order)])

    basis = R

    uvw = np.dot(basis, xyz) + origin[:, np.newaxis]
    u, v, w = uvw[0, :].reshape(orig_shape, order='C'), uvw[1, :].reshape(orig_shape, order='C'), uvw[2, :].reshape(orig_shape, order='C')
    return u, v, w








def bt_rate_maxwell_average(sigma_fun, Ti, E_beam, m_bckg, m_beam, n_level):
    """Calculates Maxwellian reaction rate for a beam with atomic mass "m_beam", 
    energy "E_beam", firing into a target with atomic mass "m_bckg" and temperature "T".

    The "sigma_fun" argument must be a function for a specific charge and n-level of the beam particles.
    Ref: FIDASIM atomic_tables.f90 bt_maxwellian_n_m.

    Args:
        sigma_fun: :py:meth
            Function to compute a specific cross section [:math:`cm^2`], function of energy/amu ONLY.
            Expected call form: sigma_fun(erel/ared)
        Ti : float, 1D or 2D array
            Target temperature [keV]. Results will be computed for each Ti value in a vectorized manner.
        E_beam : float
            Beam energy [keV]
        m_bckg : float
            Target atomic mass [amu]
        m_beam : float
            Beam atomic mass [amu]
        n_level :int
            n-level of beam. This is used to evaluate the hydrogen ionization potential,
            below which an electron is unlikely to charge exchange with surrounding ions.

    Returns:
        rate : output reaction rate in [cm^3/s] units
    """
    from scipy import constants as consts

    # enforce expected shape
    Ti = np.atleast_2d(Ti)

    # radial and parallel velocity grids, in units of thermal velocity
    vr = np.linspace(0.0, 4.0, 30)
    vz = np.linspace(-4, 4.0, 60)

    # normalized energy and temperature
    E_beam_per_amu = E_beam / m_beam  # E_bar
    T_per_amu = np.maximum(Ti, 1.0e-6) / m_bckg  # T_bar

    # beam/target reduced mass:
    ared = m_bckg * m_beam / (m_bckg + m_beam)
    dE = (13.6e-3) / (n_level ** 2)  # hydrogen ionization potential

    v_therm = np.sqrt(2.0 * T_per_amu * 1.0e3 * consts.e / consts.m_p) * 1e2

    zb = np.sqrt(E_beam_per_amu / T_per_amu)  # sqrt(E_bar/T_bar)
    u2_to_erel = ared * T_per_amu

    if ared <= 0.5:  # for electron interactions
        ared = 1.0

    fr = np.zeros((Ti.shape[0], Ti.shape[1], len(vr)))
    fz = np.zeros((Ti.shape[0], Ti.shape[1], len(vz)))

    for i in np.arange(len(vz)):
        for j in np.arange(len(vr)):
            # relative square velocity:
            u2 = (zb - vz[i]) ** 2 + vr[j] ** 2
            Erel = u2_to_erel * u2

            sig = np.zeros_like(Erel)
            mask = Erel >= dE  # no possible interaction below hydrogen ionization potential
            sig[mask] = sigma_fun(Erel[mask] / ared)

            fr[:, :, j] = sig * np.sqrt(u2) * np.exp(-(vz[i] ** 2.0 + vr[j] ** 2.0)) * vr[j]

        fz[:, :, i] = scipy.integrate.simps(fr, vr, axis=-1)

    # effective maxwellian-averaged rate:
    sig_eff = (2.0 / np.sqrt(np.pi)) * scipy.integrate.simps(fz, vz, axis=-1)
    rate = sig_eff * v_therm

    return rate




def tt_rate_maxwell_average(sigma_fun, Ti, m_i, m_n, n_level):
    """Calculates Maxwellian reaction rate for an interaction between two thermal populations,
    assumed to be of neutrals (mass m_n) and background ions (mass m_i).

    The 'sigma_fun' argument must be a function for a specific charge and n-level of the neutral 
    particles. This allows evaluation of atomic rates for charge exchange interactions between thermal
    beam halos and background ions.

    Args:
        sigma_fun: python function
            Function to compute a specific cross section [cm^2], function of energy/amu ONLY.
            Expected call form: sigma_fun(erel/ared)
        Ti: float or 1D array
            background ion and halo temperature [keV]
        m_i: float
            mass of background ions [amu]
        m_n: float 
            mass of neutrals [amu]
        n_level: int
            n-level of beam. This is used to evaluate the hydrogen ionization potential,
            below which an electron is unlikely to charge exchange with surrounding ions.

        TODO: add effect of toroidal rotation! This will require making the integration in this
        function 2-dimensional.

    Returns:
        rate : float or 1D array
            output reaction rate in [cm^3/s] units
    """
    Ti = np.atleast_1d(Ti)

    vz = np.linspace(0, 4.0, 60)
    Erel = Ti[:, np.newaxis] * vz[np.newaxis, :] ** 2

    # normalized energy and temperature
    T_per_amu = np.maximum(Ti, 1.0e-6) / m_n  # T_bar

    integrand = (
        lambda erel: bt_rate_maxwell_average(sigma_fun, Ti, erel, m_i, m_n, n_level) * (2.0 * m_n * erel) ** (-0.5) * np.exp(-erel / Ti)
    )

    dE = (13.6e-3) / (n_level ** 2)  # hydrogen ionization potential
    sigma = np.zeros_like(Erel)

    for ie in np.arange(len(vz)):  # loop over vz
        mask = Erel[:, ie] >= dE  # no possible interaction below hydrogen ionization potential
        sigma[mask, ie] = bt_rate_maxwell_average(sigma_fun, Ti[mask], Erel[mask, ie], m_i, m_n, n_level)

    prefactor = np.sqrt(2.0 / (np.pi * T_per_amu)) ** (-0.5)
    sigmav = prefactor * scipy.integrate.simps(sigma, Erel, axis=-1)

    return sigmav




import numpy as np

def load_default_namelist():
    ''' Load default namelist. 
    Users should modify and complement this for a successful forward-model run.
    '''

    namelist = {
        'device': 'CMOD',
        'shot': 99999,
        'time': 1250, # in ms, for equilibrium
        'imp': 'Ca',
        # --------------------
        # background:
        'main_element': 'D',
        'Baxis': 5.5,
        # --------------------
        # explicit source:
        'explicit_source_vals' : None, # provide 2D array on explicit_source_time and explicit_source_rhop grids
        'explicit_source_time' : None, 
        'explicit_source_rhop' : None,
        # other options to specify source:
        'source_width_in' :  0.0,   # exponential ionization decay from wall boundary if widths both = 0
        'source_width_out': 0.0,    
        'imp_source_energy_eV' : 3.0, # only needed if source_width_in=source_width_out=0.0
        'prompt_redep_flag' : False,
        'Phi0': 1e+21,
        'source_type':'const', 
        'source_file': None, # required if source_type='file'
        'source_cm_out_lcfs' : 1.0, # source distance in cm from LCFS
        # LBO synthetic model, only used if source_type='synth_LBO'
        'LBO': {'n_particles': 1e+18,
                't_fall': 1.71,
                't_rise': 0.181,
                't_start': 0.0},
        # --------------------
        # timing
        'timing': {'dt_increase': np.array([1.005, 1.]),
                   'dt_start': np.array([1e-5, 0.001]),
                   'steps_per_cycle': np.array([1, 1]),
                   'times': np.array([0., 0.1])},
        # --------------------
        # radial grid
        'K': 6.0,
        'bound_sep': 2.0,
        'clen_divertor': 17.0,
        'clen_limiter': 0.5,
        'dr_0': 0.3,
        'dr_1': 0.05,
        'lim_sep': 1.0,
        'SOL_decay': 0.05, 
        # -------------------
        'saw_model': {'saw_flag': False,
                      'rmix': 1000.0,
                      'times': [1.0,], 
                      'sawtooth_erfc_width' : 1.0},
        # --------------------
        # edge/recycling
        'recycling_flag' : False,
        'wall_recycling' : 0.0, 
        'divbls' : 0.0,
        'tau_div_SOL_ms' : 50.0,  # ms
        'tau_pump_ms' : 500.0,  # ms
        'tau_rcl_ret_ms' : 50.0,   # ms
        'SOL_mach': 0.1,
        # --------------------
        # kinetic profiles
        'average_kin_profs': True,
        'kin_profs' : {'ne': {'fun': 'interpa',
                              'times': [1.]}, 
                       'Te': {'fun':'interp',
                              'times': [1.], 'decay': [1.]},
                       'Ti': {'fun':'interp',
                              'times': [1.], 'decay': [1.]},
                       'n0': {'fun':'interpa',
                              'times': [1.]}
                   },
        # ----------------
        # flags
        'cxr_flag' : False,
        'nbi_cxr_flag' : False,
        'prad_flag' : False,
        'thermal_cx_rad_flag' : False,
        'spectral_brem_flag' : False,
        'sxr_flag' : False,
        'main_ion_brem_flag' : False,
    }
    return namelist



'''
Methods related to impurity source functions.

sciortino, 2020
'''

import numpy as np
import copy,sys
from scipy.constants import m_p, e as q_electron
from scipy.special import erfc

def get_source_time_history(namelist, Raxis_cm, time):
    '''Load source time history based on current state of the namelist.

    There are 4 options to describe the time-dependence of the source:

    (1) namelist['source_type'] == 'file': in this case, a simply formatted 
    source file, with one time point and corresponding and source amplitude on each
    line, is read in. This can describe an arbitrary time dependence, e.g. 
    as measured from an experimental diagnostic. 

    (2) namelist['source_type'] == 'const': in this case, a constant source 
    (e.g. a gas puff) is simulated. It is recommended to run the simulation for 
    >100ms in order to see self-similar charge state profiles in time. 

    (3) namelist['source_type'] == 'step': this allows the creation of a source
    that suddenly appears and suddenly stops, i.e. a rectangular "step". The 
    duration of this step is given by namelist['step_source_duration']. Multiple 
    step times can be given as a list in namelist['src_step_times']; the amplitude
    of the source at each step is given in namelist['src_step_rates']

    (4) namelist['source_type'] == 'synth_LBO': this produces a model source from a LBO
    injection, given by a convolution of a gaussian and an exponential. The required 
    parameters in this case are inside a namelist['LBO'] dictionary:
    namelist['LBO']['t_start'], namelist['LBO']['t_rise'], namelist['LBO']['t_fall'], 
    namelist['LBO']['n_particles']. The "n_particles" parameter corresponds to the 
    amplitude of the source (the number of particles corresponding to the integral over
    the source function. 

    Args:
        namelist : dict
            Aurora namelist dictionary.
        Raxis_cm : float
            Major radius at the magnetic axis [cm]. This is needed to normalize the 
            source such that it is treated as toroidally symmetric -- a necessary
            idealization for 1.5D simulations. 
        time : array (nt,), optional
            Time array the source should be returned on.

    Returns: 
        source_time_history : array (nt,)
            The source time history on the input time base.
    '''
    imp = namelist['imp']

    if namelist['source_type'] == 'file':
        src_times, src_rates = read_source(namelist['source_file'])  

    elif namelist['source_type'] == 'const':
        src_times = copy.deepcopy(time)
        src_rates = np.ones(len(time)) * namelist['Phi0']
        src_rates[0] = 0.0 # start with 0

    elif namelist['source_type'] == 'step':
        # Create the time-dependent step source
        tbuf = namelist['step_source_duration']   # e.g. 1.e-6

        # Start with zero source at t=0
        src_times = [0.]
        src_rates = [0.]

        # Form the source with "tbuf" duration to connect the steps
        for i in range(len(namelist['src_step_times'])):
            src_times.append(namelist['src_step_times'][i]-tbuf)
            src_times.append(namelist['src_step_times'][i])

            src_rates.append(src_rates[-1])
            src_rates.append(namelist['src_step_rates'][i])

        # Append the final rate to maximum time
        src_rates.append(src_rates[-1])
        src_times.append(np.max(time))

    elif namelist['source_type'] == 'synth_LBO':
        # use idealized source function
        src_times, src_rates = lbo_source_function(namelist['LBO']['t_start'],
                                                   namelist['LBO']['t_rise'], namelist['LBO']['t_fall'],
                                                   namelist['LBO']['n_particles'])
    else:
        raise ValueError('Unspecified source function time history!')

    source = np.interp(time,src_times, src_rates, left=0,right=0)
    circ = 2*np.pi*Raxis_cm   #cm

    # number of particles per cm and sec
    source_time_history = np.r_[source[1:],0]/circ #NOTE source in STRAHL is by one timestep shifted

    return np.asfortranarray(source_time_history)



def write_source(t, s, shot, imp='Ca'):
    """Write a STRAHL source file. 
        
    This will overwrite any {imp}flx{shot}.dat locally. 
        
    Args:
        t : array of float, (`n`,)
            The timebase (in seconds).
        s : array of float, (`n`,)
            The source function (in particles/s).
        shot : int
            Shot number, only used for saving to a .dat file
        imp : str, optional
            Impurity species atomic symbol

    Returns:
        contents : str
            Content of the source file written to {imp}flx{shot}.dat

    """
    contents = '{.d}\n'.format(len(t),)
    for tv, sv in zip(t, s):
        contents += '    {5.5f}    {5.5e}\n'.format(tv, sv)
        
    with open(f'{imp}flx{shot}.dat', 'w') as f:
        f.write(contents)

    return contents

def read_source(filename):
    '''Read a STRAHL source file from {imp}flx{shot}.dat locally. 
    
    Args:
        filename : str 
            Location of the file containing the STRAHL source file. 
    
    Returns:
        t : array of float, (`n`,)
            The timebase (in seconds).
        s : array of float, (`n`,)
            The source function (#/s).
    '''
    t = []; s = []
    with open(filename,'r') as f:
        lines = f.read()
    for line in lines.split('\n')[1:]:  # first line contains number of lines
        line = line.strip()
        if line != '':
            t.append(float(line.split()[0]))
            s.append(float(line.split()[1]))
    t = np.array(t)
    s = np.array(s)

    return t,s
        
    

def lbo_source_function(t_start, t_rise, t_fall, n_particles=1.0, time_vec=None):
    ''' Model for the expected shape of the time-dependent source function,
    using a convolution of a gaussian and an exponential decay.

    Args:
        t_start : float or array-like [ms]
            Injection time, beginning of source rise. If multiple values are given, they are
            used to create multiple source functions.
        t_rise : float or array-like [ms]
            Time scale of source rise. Similarly to t_start for multiple values.
        t_fall : float or array-like [ms]
            Time scale of source decay.Similarly to t_start for multiple values.
        n_particles : float, opt
            Total number of particles in source. Similarly to t_start for multiple values.
            Defaults to 1.0.
        time_vec : array-like
            Time vector on which to create source function. If left to None,
            use a linearly spaced time vector including the main features of the function.

    Returns:
        time_vec : array
            Times for the source function of each given impurity
        source : array
            Time history of the synthetized source function.
    '''

    t_start = np.atleast_1d(t_start)
    t_start,t_fall,t_rise,n_particles = np.broadcast_arrays(t_start,t_fall/1e3,t_rise/1e3,n_particles)


    if time_vec is None:
        time_vec = np.hstack([np.linspace(ts-3*tr,ts+6*tf,200) for ts,tr,tf in zip(t_start,t_rise,t_fall)])

    source = np.zeros_like(time_vec)

    for ts,tr,tf,N in zip(t_start,t_rise,t_fall,n_particles):
        tind = slice(*time_vec.searchsorted([ts-3*tr,ts+6*tf]))
        T = time_vec[tind]
        source[tind] = np.exp((1 - 4* tf*(T-ts)/tr**2)/(4*(tf/tr)**2))*(erfc((T-ts)/tr - 1/(2*tf/tr)) - 2)

        # scale source to correspond to the given total number of particles
        source[tind]*= N/np.trapz(source[tind],T)

        # ensure that source function ends with 0 to avoid numerical issues
        source[tind][[0,-1]] = 0


    return time_vec, source




def get_radial_source(namelist, rvol_grid, pro_grid, S_rates, Ti_eV=None):
    '''Obtain spatial dependence of source function.

    If namelist['source_width_in']==0 and namelist['source_width_out']==0, the source
    radial profile is defined as an exponential decay due to ionization of neutrals. This requires
    S_rates, the ionization rate of neutral impurities, to be given with S_rates.shape=(len(rvol_grid),len(time_grid))

    If namelist['imp_source_energy_eV']<0, the neutrals speed is taken as the thermal speed based
    on Ti_eV, otherwise the value corresponding to namelist['imp_source_energy_eV'] is used.

    Args:
        namelist : dict
            Aurora namelist. Only elements referring to the spatial distribution and energy of 
            source atoms are accessed. 
        rvol_grid : array (nr,)
            Radial grid in volume-normalized coordinates [cm]
        pro_grid : array (nr,)
            Normalized first derivatives of the radial grid in volume-normalized coordinates. 
        S_rates : array (nr,nt)
            Ionization rate of neutral impurity over space and time.

    Keyword Args:
        Ti_eV : array, optional (nr,nt)
            Background ion temperature, only used if source_width_in=source_width_out=0.0 and 
            imp_source_energy_eV<=0, in which case the source impurity neutrals are taken to 
            have energy equal to the local Ti [eV]. 

    Returns:
        source_rad_prof : array (nr,nt)
            Radial profile of the impurity neutral source for each time step.
    '''
    r_src = namelist['rvol_lcfs'] + namelist['source_cm_out_lcfs']

    source_rad_prof = np.zeros_like(S_rates)

    # find index of radial grid vector that is just greater than r_src
    i_src = rvol_grid.searchsorted(r_src)-1
    # set source to be inside of the wall
    i_src = min(i_src, len(rvol_grid)-1)

    if (namelist['source_width_in'] < 0. and namelist['source_width_out'] < 0.):
        # point source
        source_rad_prof[i_src] = 1.0

    # source with FWHM=source_width_in inside and FWHM=source_width_out outside
    if namelist['source_width_in']>0. or namelist['source_width_out']>0.:

        if namelist['source_width_in']> 0.:
            ff = np.log(2.)/namelist['source_width_in']**2
            source_rad_prof[:i_src] = np.exp(-(rvol_grid[:i_src]-rvol_grid[i_src])**2*ff)[:,None]

        if namelist['source_width_out']>0.:
            ff = np.log(2.)/namelist['source_width_out']**2
            source_rad_prof[i_src:] = np.exp(-(rvol_grid[i_src:]-rvol_grid[i_src])**2*ff)[:,None]


    # decay of neutral density with exp(-Int( [ne*S+dv/dr]/v ))
    if namelist['source_width_in']==0 and namelist['source_width_out']==0:
        #neutrals energy
        if namelist['imp_source_energy_eV'] > 0: 
            E0 = namelist['imp_source_energy_eV']*np.ones_like(rvol_grid)
        else:
            if Ti_eV is not None:
                E0 = copy.deepcopy(Ti_eV)
            else:
                raise ValueError('Could not compute a valid energy of injected ions!')

        # import here to avoid issues with omfit_commonclasses during docs and package creation
        from omfit_commonclasses.utils_math import atomic_element

        # velocity of neutrals [cm/s]
        out = atomic_element(symbol=namelist['main_element'])
        spec = list(out.keys())[0]
        main_ion_A = int(out[spec]['A'])
        v = - np.sqrt(2.*q_electron*E0/(main_ion_A*m_p))

        #integration of ne*S for atoms and calculation of ionization length for normalizing neutral density
        source_rad_prof[i_src]= -0.0625*S_rates[i_src]/pro_grid[i_src]/v[i_src]   #1/16
        for i in np.arange(i_src-1,0,-1):
            source_rad_prof[i]=source_rad_prof[i+1]+0.25*(
                S_rates[i+1]/pro_grid[i+1]/v[i+1] + S_rates[i]/pro_grid[i]/v[i])

        #prevents FloatingPointError: underflow encountered
        source_rad_prof[1:i_src] = np.exp(np.maximum(source_rad_prof[1:i_src],-100))

        # calculate relative density of neutrals
        source_rad_prof[1:i_src] *= (rvol_grid[i_src]*v[i_src]/rvol_grid[1:i_src]/v[1:i_src])[:,None]
        source_rad_prof[i_src]=1.0

        # remove promptly redeposited ions
        if namelist['prompt_redep_flag']:
            omega_c = 1.602e-19/1.601e-27*namelist['Baxis']/namelist['a']  
            dt = (rvol_grid[i_src]-rvol_grid)/v
            pp = dt*omega_c
            non_redep =  pp**2/(1.+pp**2)
            source_rad_prof*=  non_redep


    # total ion source
    pnorm = np.pi*np.sum(source_rad_prof*S_rates*(rvol_grid/pro_grid)[:,None],0)  # sum over radius
    
    # neutral density for influx/unit-length = 1/cm
    source_rad_prof /= pnorm

    return np.asfortranarray(source_rad_prof)




'''
Script to demonstrate the creation of an animation of an Aurora run, purely for visualization purposes. 

It is recommended to run this in basic Python3 (not IPython or notebooks).
Note that you might need to run %matplotlib qt in IPython in order to enable the animation to run. 
'''

import numpy as np
import matplotlib.pyplot as plt; plt.ion()
import omfit_eqdsk, omfit_gapy
import pickle as pkl
import scipy,sys,os
import time

# Make sure that package home is added to sys.path
sys.path.append('../')
import aurora


# read in default Aurora namelist
namelist = aurora.default_nml.load_default_namelist()
kp = namelist['kin_profs']

# Use gfile and statefile in local directory:
examples_dir = os.path.dirname(os.path.abspath(__file__))
geqdsk = omfit_eqdsk.OMFITgeqdsk(examples_dir+'/example.gfile')
inputgacode = omfit_gapy.OMFITgacode(examples_dir+'/example.input.gacode')

# transform rho_phi (=sqrt toroidal flux) into rho_psi (=sqrt poloidal flux) and save kinetic profiles
kp['Te']['rhop'] = kp['ne']['rhop'] = np.sqrt(inputgacode['polflux']/inputgacode['polflux'][-1])
kp['ne']['vals'] = inputgacode['ne'][None,:]*1e13 # 1e19 m^-3 --> cm^-3
kp['Te']['vals'] = inputgacode['Te'][None,:]*1e3  # keV --> eV

# set impurity species and sources rate
imp = namelist['imp'] = 'Ar'
namelist['source_type'] = 'const'
namelist['Phi0'] = 2e20 # particles/s

# Now get aurora setup
asim = aurora.core.aurora_sim(namelist, geqdsk=geqdsk)

# set time-independent transport coefficients (flat D=1 m^2/s, V=-2 cm/s)
D_z = 1e4 * np.ones(len(asim.rvol_grid))  # cm^2/s
V_z = -2e2 * np.ones(len(asim.rvol_grid)) # cm/s

# run Aurora forward model
out = asim.run_aurora(D_z, V_z)

# extract densities of each charge state:
nz = out[0]

# now create animation
aurora.animate.animate_aurora(asim.rhop_grid, asim.time_out, nz.transpose(1,0,2),
                              xlabel=r'$\rho_p$', ylabel='t={:.4f} [s]', zlabel=r'$n_z$ [$cm^{-3}$]',
                              labels=[str(i) for i in np.arange(0,nz.shape[1])],
                              plot_sum=True, save_filename='aurora_anim')



import os,sys
import numpy as np
from scipy.interpolate import RectBivariateSpline, interp1d
from scipy.integrate import cumtrapz
import matplotlib.pyplot as plt
plt.ion()
from scipy import constants
import warnings, copy

from . import atomic
from . import adas_files
from . import plot_tools


def compute_rad(imp, nz, ne, Te,
                n0 = None, Ti = None, ni = None, adas_files_sub = {},
                prad_flag=False, sxr_flag=False,
                thermal_cx_rad_flag=False, spectral_brem_flag=False, ):
    '''Calculate radiation terms corresponding to a simulation result. The nz,ne,Te,n0,Ti,ni arrays
    are normally assumed to be given as a function of (time,nZ,space), but time and space may 
    be substituted by other coordinates (e.g. R,Z)
    
    Result can be conveniently plotted with a time-slider using, for example

    .. code-block:: python

        aurora.slider_plot(rhop,time, res['line_rad'].transpose(1,2,0)/1e6,
            xlabel=r'$\\rho_p$', ylabel='time [s]', 
            zlabel=r'$P_{rad}$ [$MW$]',
            plot_sum=True,
            labels=[f'Ca$^{{{str(i)}}}$' for i in np.arange(res['line_rad'].shape[1])])

    All radiation outputs are given in :math:`W cm^{-3}`, consistently with units of :math:`cm^{-3}`
    given for inputs.

    Args:
        imp : str
             Impurity symbol, e.g. Ca, F, W
        nz : array (time, nZ, space) [:math:`cm^{-3}`]
            Dictionary with impurity density result, as given by :py:func:`~aurora.core.run_aurora` method.
        ne : array (time,space) [:math:`cm^{-3}`]
            Electron density on the output grids.
        Te : array (time,space) [eV]
            Electron temperature on the output grids.


    Keyword Args:
        n0 : array(time,space), optional [:math:`cm^{-3}`]
             Background neutral density (assumed of hydrogen-isotopes).
             This is only used if thermal_cx_rad_flag=True.
        Ti : array (time,space) [eV]
            Main ion temperature (assumed of hydrogen-isotopes). This is only used
            if thermal_cx_rad_flag=True. If not set, Ti is taken equal to Te. 
        adas_files_sub : dict
            Dictionary containing ADAS file names for radiation calculations, possibly including keys
            "plt","prb","prc","pls","prs","pbs","brs"
            Any file names that are needed and not provided will be searched in the 
            :py:meth:`~aurora.adas_files.adas_files_dict` dictionary. 
        prad_flag : bool, optional
            If True, total radiation is computed (for each charge state and their sum)
        sxr_flag : bool, optional
            If True, soft x-ray radiation is computed (for the given 'pls','prs' ADAS files)
        thermal_cx_rad_flag : bool, optional
            If True, thermal charge exchange radiation is computed.
        spectral_brem_flag : bool, optional
            If True, spectral bremstrahlung is computed (based on available 'brs' ADAS file)

    Returns:
        res : dict
            Dictionary containing radiation terms, depending on the activated flags. 
            The structure of the "res" dictionary is as follows.

        If prad_flag=True,

        res['line_rad'] : array (nt,nZ,nr)- from ADAS "plt" files
            Excitation-driven line radiation for each impurity charge state.
        res['cont_rad'] : array (nt,nZ,nr)- from ADAS "prb" files
            Continuum and line power driven by recombination and bremsstrahlung for impurity ions.
        res['brems'] : array (nt,nr)- analytic formula. 
            Bremsstrahlung produced by electron scarrering at fully ionized impurity 
            This is only an approximate calculation and is more accurately accounted for in the 
            'cont_rad' component.
        res['thermal_cx_cont_rad'] : array (nt,nZ,nr)- from ADAS "prc" files
            Radiation deriving from charge transfer from thermal neutral hydrogen to impurity ions.
            Returned only if thermal_cx_rad_flag=True.
        res['tot'] : array (nt,nZ,nr)
            Total unfilted radiation, summed over all charge states, given by the sum of all known 
            radiation components.

        If sxr_flag=True,

        res['sxr_line_rad'] : array (nt,nZ,nr)- from ADAS "pls" files
            Excitation-driven line radiation for each impurity charge state in the SXR range.
        res['sxr_cont_rad'] : array (nt,nZ,nr)- from ADAS "prs" files
            Continuum and line power driven by recombination and bremsstrahlung for impurity ions
            in the SXR range. 
        res['sxr_brems'] : array (nt,nZ,nr)- from ADAS "pbs" files
            Bremsstrahlung produced by electron scarrering at fully ionized impurity in the SXR range.
        res['sxr_tot'] : array (nt,nZ,nr)
            Total radiation in the SXR range, summed over all charge states, given by the sum of all known 
            radiation components in the SXR range. 

        If spectral_brem_flag,

        res['spectral_brems'] : array (nt,nZ,nr) -- from ADAS "brs" files
            Bremsstrahlung at a specific wavelength, depending on provided "brs" file. 
    '''
    res = {}

    Z_imp = nz.shape[1] - 1
    logTe = np.log10(Te)
    logne = np.log10(ne)

    # calculate total radiation
    if prad_flag:

        if 'plt' in adas_files_sub:  # check if user requested use of a specific file
            atom_data = atomic.get_atom_data(imp, ['plt'],[adas_files_sub['plt']])
        else:  # use default file from adas_files.adas_files_dict()
            atom_data = atomic.get_atom_data(imp, ['plt'])
        pltt = atomic.interp_atom_prof(atom_data['plt'],logne,logTe) # W

        if 'prb' in adas_files_sub:
            atom_data = atomic.get_atom_data(imp, ['prb'],[adas_files_sub['prb']])
        else:
            atom_data = atomic.get_atom_data(imp, ['prb'])
        prb = atomic.interp_atom_prof(atom_data['prb'],logne,logTe) # W

        # line radiation for each charge state
        res['line_rad'] = np.maximum(nz[:,:-1] * pltt, 1e-60) # no line rad for fully stripped ion       

        # total continuum radiation (NB: neutrals do not have continuum radiation)
        res['cont_rad'] = nz[:,1:] * prb

        # impurity brems (inaccurate Gaunt factor!) -- already included in 'cont_rad'
        res['brems'] = atomic.impurity_brems(nz, ne, Te)

        # Total unfiltered radiation: 
        res['tot'] = res['line_rad'].sum(1) + res['cont_rad'].sum(1) 

        if thermal_cx_rad_flag:
            if n0 is None:
                raise ValueError(
                    'Requested thermal CX emission to be computed, '
                    'but no background neutral density was provided!')
            if Ti is None:
                warnings.warn('Requested thermal CX emission to be computed '
                              'but no Ti values were provided! Setting Ti=Te',
                              RuntimeWarning)
                Ti = copy.deepcopy(Te)

            # make sure that n0 and Ti are given as 2D:
            assert n0.ndim==2 and Ti.ndim==2
            
            logTi = np.log10(Ti)
            
            # thermal CX radiation to total recombination and continuum radiation terms:
            if 'prc' in adas_files_sub:
                atom_data = atomic.get_atom_data(imp, ['prc'],[adas_files_sub['prc']])
            else:
                atom_data = atomic.get_atom_data(imp, ['prc'])

            # prc has weak dependence on density, so no difference between using ni or ne
            prc = atomic.interp_atom_prof(atom_data['prc'],logne,logTi,x_multiply=False) # W

            # broadcast n0 to dimensions (nt,nZ,nr):
            res['thermal_cx_cont_rad'] = nz[:,1:] * n0[:,None] * prc

            # add to total unfiltered radiation:
            res['tot'] += res['thermal_cx_cont_rad'].sum(1)
                       
    if sxr_flag: # SXR-filtered radiation (spectral range depends on filter used for files)

        if 'pls' in adas_files_sub:
            atom_data = atomic.get_atom_data(imp, ['pls'],[adas_files_sub['pls']])
        else:
            atom_data = atomic.get_atom_data(imp, ['pls'])
        pls = atomic.interp_atom_prof(atom_data['pls'],logne,logTe) # W

        if 'prs' in adas_files_sub:
            atom_data = atomic.get_atom_data(imp, ['prs'],[adas_files_sub['prs']])
        else:
            atom_data = atomic.get_atom_data(imp, ['prs'])
        prs = atomic.interp_atom_prof(atom_data['prs'],logne,logTe) # W

        # SXR line radiation for each charge state
        res['sxr_line_rad'] = np.maximum(nz[:,:-1] * pls, 1e-60)

        # SXR continuum radiation for each charge state
        res['sxr_cont_rad'] = nz[:,1:] * prs

        try:
            # impurity bremsstrahlung in SXR range -- already included in 'sxr_cont_rad'
            if 'pbs' in adas_files_sub:
                atom_data = atomic.get_atom_data(imp, ['pbs'],[adas_files_sub['pbs']])
            else:
                atom_data = atomic.get_atom_data(imp, ['pbs'])
            pbs = atomic.interp_atom_prof(atom_data['pbs'],logne,logTe) # W
            res['sxr_brems'] = nz[:,1:] * pbs 
        except IndexError:
            # pbs file not available by default for this ion. Users may specify it in adas_files_sub
            pass
        
        # SXR total radiation
        res['sxr_tot'] = res['sxr_line_rad'].sum(1) + res['sxr_cont_rad'].sum(1)

        
    if spectral_brem_flag:  # spectral bremsstrahlung (i.e. brems at a specific wavelength)

        if 'brs' in adas_files_sub:
            atom_data = atomic.get_atom_data(imp, ['brs'],[adas_files_sub['brs']])
        else:
            atom_data = atomic.get_atom_data(imp, ['brs'])
        x,y,tab = atom_data['brs']
        brs = atomic.interp_atom_prof((x,y,tab.T),None,logTe) # W

        # interpolate on Z grid of impurity of interest
        logZ_rep = np.log10(np.arange(Z_imp)+1)
        brs = interp1d(x, brs,axis=1,copy=False,assume_sorted=True)(logZ_rep)

        # Note: no spectral bremsstrahlung from neutral stage
        res['spectral_brems'] = nz[:,1:] * brs

    return res



def plot_radiation_profs(imp, nz_prof, logne_prof, logTe_prof, xvar_prof,
                         xvar_label='', atom_data=None):
    '''Compute profiles of predicted radiation, both SXR-filtered and unfiltered.
    This function offers a simplified interface to radiation calculation with respect to 
    :py:meth:`~aurora.radiation.compute_rad`, which is more complete.

    This function can be used to plot radial profiles (setting xvar_prof to a radial grid)
    or profiles as a function of any variable on which the logne_prof and logTe_prof
    may depend.

    The variable "nz_prof" may be a full description of impurity charge state densities
    (e.g. the output of aurora), or profiles of fractional abundances from ionization 
    equilibrium.

    Args: 
        imp : str, optional
            Impurity ion atomic symbol.
        nz_prof : array (TODO for docs: check dimensions)
            Impurity charge state densities
        logne_prof : array (TODO for docs: check dimensions)
            Electron density profiles in :math:`cm^{-3}`
        logTe_prof : array (TODO for docs: check dimensions)
            Electron temperature profiles in eV
        xvar_prof : array (TODO for docs: check dimensions)
            Profiles of a variable of interest, on the same grid as kinetic profiles. 
        xvar_label : str, optional
            Label for x-axis. 
        atom_data : dict, optional
            Dictionary containing atomic data as output by :py:meth:`~aurora.atomic.get_atom_data`
            for the atomic processes of interest. "prs","pls","plt" and "prb" are required by this function.
            If not provided, this function loads these files internally. 

    Returns:
        pls : array (TODO for docs: check dimensions)
            SXR line radiation.
        prs : array (TODO for docs: check dimensions)
            SXR continuum radiation.
        pltt : array (TODO for docs: check dimensions)
            Unfiltered line radiation.
        prb : array (TODO for docs: check dimensions)
            Unfiltered continuum radiation.        
    '''
    if atom_data is None:
        # if atom_data dictionary was not given, load appropriate default files
        atom_data = atomic.get_atom_data(imp,['pls','prs','plt','prb'])

    # use "pltt" nomenclature rather than "plt" to avoid issues with matplotlib.pyplot imported as plt
    pls, prs, pltt, prb = atomic.get_cooling_factors(atom_data, logTe_prof, nz_prof, ion_resolved = True, plot=False)

    emiss_sxr = np.zeros((len(xvar_prof),nion))
    emiss_tot = np.zeros((len(xvar_prof),nion))
    emiss_sxr[:, 1: ] += prs
    emiss_sxr[:, :-1] += pls
    emiss_tot[:, 1: ] += prb
    emiss_tot[:, :-1] += pltt

    # plot radiation components
    fig,axx = plt.subplots(2,2,figsize=(12,8),sharex=True)
    ax = axx.flatten()
    nion = prs.shape[1]+1
    colors = cm.plasma(np.linspace(0,1, nion))
    for a in ax:
        a.set_prop_cycle('color',colors)
        a.grid(True)

    ax[0].plot([],[]) #empty plot for bremstrahlung of neutral ion
    ax[0].plot(xvar_prof,prs); ax[0].set_title('PRS: cont SXR rad')
    ax[1].plot(xvar_prof,pls); ax[1].set_title('PLS: SXR line rad')
    
    ax[2].plot(xvar_prof,pltt); ax[2].set_title('PLT: tot line rad')
    ax[3].plot([],[]) #empty plot for bremstrahlung of neutral ion
    ax[3].plot(xvar_prof,prb); ax[3].set_title('PRB: tot cont rad')

    ax[2].set_xlabel(xvar_label)
    ax[3].set_xlabel(xvar_label)
    
    labels = [r'$%s^{%d\!+}$'%(imp,cc) for cc in range(nion)]
    ax[0].legend(labels)
    ax[0].set_xlim(xvar_prof[0], xvar_prof[-1])

    # plot total power (in SXR and whole range)
    fig,axx = plt.subplots(2,2,figsize=(12,8),sharex=True)
    ax = axx.flatten()

    for a in ax:
        a.set_prop_cycle('color',colors)
        a.grid(True)

    ax[0].plot(xvar_prof,emiss_sxr); ax[0].set_title('SXR power [W]')
    ax[1].plot(xvar_prof,emiss_tot); ax[1].set_title('Tot. rad. power [W]')
    ax[2].plot(xvar_prof,emiss_sxr*10**logne_prof[:,None]); ax[2].set_title(r'SXR power [W/m$^{-3}$]')
    ax[3].plot(xvar_prof,emiss_tot*10**logne_prof[:,None]); ax[3].set_title(r'Tot. rad. power [W/m$^{-3}$]')
    ax[2].set_xlabel(xvar_label)
    ax[3].set_xlabel(xvar_label)
    ax[0].legend(labels)
    ax[0].set_xlim(xvar_prof[0], xvar_prof[-1])

    return pls, prs, pltt, prb



def radiation_model(imp,rhop, ne_cm3, Te_eV, vol,
                    adas_files_sub={}, n0_cm3=None, Ti_eV=None, nz_cm3=None, frac=None, plot=False):
    '''Model radiation from a fixed-impurity-fraction model or from detailed impurity density
    profiles for the chosen ion. This method acts as a wrapper for :py:method:compute_rad(), 
    calculating radiation terms over the radius and integrated over the plasma cross section. 

    Args:
        imp : str (nr,)
            Impurity ion symbol, e.g. W
        rhop : array (nr,)
            Sqrt of normalized poloidal flux array from the axis outwards
        ne_cm3 : array (nr,)
            Electron density in :math:`cm^{-3}` units.
        Te_eV : array (nr,)
            Electron temperature in eV
        vol : array (nr,)
            Volume of each flux surface in :math:`m^3`. Note the units! We use :math:`m^3` here
            rather than :math:`cm^3` because it is more common to work with :math:`m^3` for 
            flux surface volumes of fusion devices.

    Keyword Args:
        adas_files_sub : dict
            Dictionary containing ADAS file names for forward modeling and/or radiation calculations.
            Possibly useful keys include
            "scd","acd","ccd","plt","prb","prc","pls","prs","pbs","brs"
            Any file names that are needed and not provided will be searched in the 
            :py:meth:`~aurora.adas_files.adas_files_dict` dictionary. 
        n0_cm3 : array (nr,), optional
            Background ion density (H,D or T). If provided, charge exchange (CX) 
            recombination is included in the calculation of charge state fractional 
            abundances.
        Ti_eV : array (nr,), optional
            Background ion density (H,D or T). This is only used if CX recombination is 
            requested, i.e. if n0_cm3 is not None. If not given, Ti is set equal to Te. 
        nz_cm3 : array (nr,nz), optional
            Impurity charge state densities in :math:`cm^{-3}` units. Fractional abundancies can 
            alternatively be specified via the :param:frac parameter for a constant-fraction
            impurity model across the radius. If provided, nz_cm3 is used. 
        frac : float, optional
            Fractional abundance, with respect to ne, of the chosen impurity. 
            The same fraction is assumed across the radial profile. If left to None,
            nz_cm3 must be given. 
        plot : bool, optional
            If True, plot a number of diagnostic figures. 

    Returns:
        res : dict
            Dictionary containing results of radiation model.     
    '''
    if nz_cm3 is None:
        assert frac is not None
    
    # limit all considerations to inside LCFS
    ne_cm3 = ne_cm3[rhop<=1.]
    Te_eV = Te_eV[rhop<=1.]
    vol = vol[rhop<=1.]
    if n0_cm3 is not None:
        n0_cm3 = n0_cm3[rhop<=1.]
    rhop = rhop[rhop<=1.]

    # create results dictionary
    out = {}
    out['rhop'] = rhop
    out['ne_cm3'] = ne_cm3
    out['Te_eV'] = Te_eV
    out['vol'] = vol
    if n0_cm3 is not None:
        out['n0_cm3'] = n0_cm3
        
    # load ionization and recombination rates
    filetypes = ['acd','scd']
    filenames = []
    def_adas_files_dict = adas_files.adas_files_dict()
    for filetype in filetypes:
        if filetype in adas_files_sub:
            filenames.append(adas_files_sub[filetype])
        else:
            filenames.append(def_adas_files_dict[imp][filetype])

    # if background neutral density was given, load thermal CX rates too
    if n0_cm3 is not None:
        filetypes.append('ccd')
        if 'ccd' in adas_files_sub:
            filenames.append(adas_files_sub['ccd'])
        else:
            filenames.append(def_adas_files_dict[imp]['ccd']) 

    if nz_cm3 is None:
        # obtain fractional abundances via a constant-fraction model 
        atom_data = atomic.get_atom_data(imp,filetypes,filenames)

        if n0_cm3 is None:
            # obtain fractional abundances without CX:
            logTe, out['fz'],rates = atomic.get_frac_abundances(atom_data,ne_cm3,Te_eV,rho=rhop, plot=plot)
        else:
            # include CX for ionization balance:
            logTe, out['fz'],rates = atomic.get_frac_abundances(atom_data,ne_cm3,Te_eV,rho=rhop, plot=plot,
                                                   include_cx=True, n0_by_ne=n0_cm3/ne_cm3)
        out['logTe'] = logTe
        
        # Impurity densities
        nz_cm3 = frac * ne_cm3[None,:,None] * out['fz'][None,:,:]  # (time,nZ,space)
    else:
        # set input nz_cm3 into the right shape for compute_rad: (time,space,nz)
        nz_cm3 = nz_cm3[None,:,:]

        # calculate fractional abundances 
        fz = nz_cm3[0,:,:].T/np.sum(nz_cm3[0,:,:],axis=1)
        out['fz'] = fz.T  # (nz,space)


    # compute radiated power components for impurity species
    rad = compute_rad(imp, nz_cm3.transpose(0,2,1), ne_cm3[None,:], Te_eV[None,:],
                      n0=n0_cm3[None,:] if n0_cm3 is not None else None,
                      Ti=Te_eV[None,:] if Ti_eV is None else Ti_eV[None,:],
                      adas_files_sub=adas_files_sub,
                      prad_flag=True, sxr_flag=False,
                      thermal_cx_rad_flag=False if n0_cm3 is None else True,
                      spectral_brem_flag=False)

    # radiation terms -- converted from W/cm^3 to W/m^3
    out['line_rad_dens'] = rad['line_rad'][0,:,:]*1e6
    out['cont_rad_dens'] = rad['cont_rad'][0,:,:]*1e6
    out['brems_dens'] = rad['brems'][0,:,:]*1e6
    out['rad_tot_dens'] = rad['tot'][0,:]*1e6
    
    # cumulative integral over all volume
    out['line_rad'] = cumtrapz(out['line_rad_dens'], vol, initial=0.)
    out['line_rad_tot'] = cumtrapz(out['line_rad_dens'].sum(0), vol, initial=0.)
    out['cont_rad'] = cumtrapz(out['cont_rad_dens'], vol, initial=0.)
    out['brems'] = cumtrapz(out['brems_dens'], vol, initial=0.)
    out['rad_tot'] = cumtrapz(out['rad_tot_dens'], vol, initial=0.)

    if n0_cm3 is not None:
        out['thermal_cx_rad_dens'] = rad['thermal_cx_cont_rad'][0,:,:]*1e6
        out['thermal_cx_rad'] = cumtrapz(out['thermal_cx_rad_dens'].sum(0), vol, initial=0.)
        
    # total power is the last element of the cumulative integral
    out['Prad'] = out['rad_tot'][-1]
    
    #print(f'Total {imp} line radiation power: {out["line_rad_tot"][-1]/1e6:.3f} MW')
    #print(f'Total {imp} continuum radiation power: {out["cont_rad"].sum(0)[-1]/1e6:.3f} MW')
    #print(f'Total {imp} bremsstrahlung radiation power: {out["brems"].sum(0)[-1]/1e6:.3f} MW')
    #if n0_cm3 is not None:
    #    print(f'Thermal CX power: {out["thermal_cx_rad"][-1]/1e6:.3f} MW')
    #print(f'Total radiated power: {out["Prad"]/1e6:.3f} MW')

    # calculate average charge state Z across radius
    out['Z_avg'] = np.sum(np.arange(out['fz'].shape[1])[:,None] * out['fz'].T, axis=0)
    
    if plot:
        # plot power in MW/m^3
        fig,ax = plt.subplots()
        ax.plot(rhop, out['line_rad_dens'].sum(0)/1e6, label=r'$P_{rad,line}$')
        ax.plot(rhop, out['cont_rad_dens'].sum(0)/1e6, label=r'$P_{cont}$')
        ax.plot(rhop, out['brems_dens'].sum(0)/1e6, label=r'$P_{brems}$')
        ax.plot(rhop, out['rad_tot_dens']/1e6, label=r'$P_{rad,tot}$')
        ax.set_xlabel(r'$\rho_p$')
        ax.set_ylabel(fr'{imp} $P_{{rad}}$ [$MW/m^3$]')
        ax.legend().set_draggable(True)
        
        # plot power in MW 
        fig,ax = plt.subplots()
        ax.plot(rhop, out['line_rad'].sum(0)/1e6, label=r'$P_{rad,line}$')
        ax.plot(rhop, out['cont_rad'].sum(0)/1e6, label=r'$P_{cont}$')
        ax.plot(rhop, out['brems'].sum(0)/1e6, label=r'$P_{brems}$')
        ax.plot(rhop, out['rad_tot']/1e6, label=r'$P_{rad,tot}$')
        ax.set_xlabel(r'$\rho_p$')
        ax.set_ylabel(fr'{imp} $P_{{rad}}$ [MW]')
        fig.suptitle('Cumulative power')
        ax.legend().set_draggable(True)
        plt.tight_layout()
        
        # plot line radiation for each charge state
        fig = plt.figure(figsize=(10,7))
        colspan = 8 if out['line_rad_dens'].shape[0]<50 else 7
        a_plot = plt.subplot2grid((10,10),(0,0),rowspan = 10, colspan = colspan, fig=fig) 
        a_legend = plt.subplot2grid((10,10),(0,8),rowspan = 10, colspan = 10-colspan, fig=fig) 
        ls_cycle = plot_tools.get_ls_cycle()
        for cs in np.arange(out['line_rad_dens'].shape[0]):
            ls = next(ls_cycle)
            a_plot.plot(rhop, out['line_rad_dens'][cs,:]/1e6, ls)
            a_legend.plot([], [], ls, label=imp+fr'$^{{{cs}+}}$')
        a_plot.set_xlabel(r'$\rho_p$')
        a_plot.set_ylabel(fr'{imp} $P_{{rad}}$ [$MW/m^3$]')
        ncol_leg = 2 if out['line_rad_dens'].shape[0]<25 else 3
        leg=a_legend.legend(loc='center right', fontsize=11, ncol=ncol_leg).set_draggable(True)
        a_legend.axis('off')
        
        # plot average Z over radius
        fig,ax = plt.subplots()
        ax.plot(rhop, out['Z_avg'])
        ax.set_xlabel(r'$\rho_p$')
        ax.set_ylabel(fr'{imp} $\langle Z \rangle$')
        plt.tight_layout()
        
    return out



def get_main_ion_dens(ne_cm3, ions, rhop_plot=None):
    '''Estimate the main ion density via quasi-neutrality. 
    This requires subtracting from ne the impurity charge state density times Z for each 
    charge state of every impurity present in the plasma in significant amounts. 
    
    Args:
        ne_cm3 : array (time,space)
            Electron density in :math:`cm^{-3}`
        ions : dict
            Dictionary with keys corresponding to the atomic symbol of each impurity under 
            consideration. The values in ions[key] should correspond to the charge state 
            densities for the selected impurity ion in :math:`cm^{-3}`, with shape 
            (time,nZ,space). 
        rhop_plot : array (space), optional
            rhop radial grid on which densities are given. If provided, plot densities of 
            all species at the last time slice over this radial grid. 

    Returns:
        ni_cm3 : array (time,space)
            Estimated main ion density in :math:`cm^{-3}`.
    '''
    ni_cm3 = copy.deepcopy(ne_cm3)

    for imp in ions:
        # extract charge state densities for given ion
        nz_cm3 = ions[imp]
        
        Z_imp = nz_cm3.shape[1] -1  # don't include neutral stage
        Z_n_imp = (np.arange(Z_imp+1)[None,:,None]*nz_cm3).sum(1)
        ni_cm3 -= Z_n_imp

    if rhop_plot is not None:
        fig,ax = plt.subplots()
        ax.plot(rhop_plot, ne_cm3[-1,:], label='electrons')
        for imp in ions:
            ax.plot(rhop_plot, ions[imp][-1,:,:].sum(0), label=imp)
        ax.set_xlabel(r'$\rho_p$')
        ax.set_ylabel(r'$cm^{-3}$')

    return ni_cm3



def read_adf15(path, order=1, plot_lines=[], ax=None, plot_3d=False):
    """Read photon emissivity coefficients from an ADAS ADF15 file.

    Returns a dictionary whose keys are the wavelengths of the lines in angstroms. 
    The value is an interpolant that will evaluate the PEC at a desired density and temperature. 

    Units for interpolation: :math:`cm^{-3}` for density; :math:`eV` for temperature.

    Args:
        path : str
            Path to adf15 file to read.
        order : int, opt
            Parameter to control the order of interpolation. Default is 1 (linear interpolation).
        
    Keyword Args:
        plot_lines : list
            List of lines whose PEC data should be displayed. Lines should be identified
            by their wavelengths. The list of available wavelengths in a given file can be retrieved
            by first running this function ones, checking dictionary keys, and then requesting a
            plot of one (or more) of them.
        ax : matplotlib axes instance
            If not None, plot on this set of axes.
        plot_3d : bool
            Display PEC data as 3D plots rather than 2D ones.

    Returns:
        pec_dict : dict
            Dictionary containing interpolation functions for each of the available lines of the
            indicated type (ionization or recombination). Each interpolation function takes as arguments
            the log-10 of ne and Te.

    Minimal Working Example (MWE)::

        filename = 'pec96#h_pju#h0.dat' # for D Ly-alpha

        # fetch file automatically, locally, from AURORA_ADAS_DIR, or directly from the web:
        path = aurora.get_adas_file_loc(filename, filetype='adf15')  

        # plot Lyman-alpha line at 1215.2 A. 
        # see available lines with pec_dict.keys() after calling without plot_lines argument
        pec_dict = aurora.read_adf15(path, plot_lines=[1215.2])

    Another example, this time also with charge exchange::

        filename = 'pec96#c_pju#c2.dat'
        path = aurora.get_adas_file_loc(filename, filetype='adf15')
        pec_dict = aurora.read_adf15(path, plot_lines=[361.7])

    Metastable-resolved files will be automatically identified and parsed accordingly, e.g.::

         filename = 'pec96#he_pjr#he0.dat'
         path = aurora.get_adas_file_loc(filename, filetype='adf15')
         pec_dict = aurora.read_adf15(path, plot_lines=[584.4])

    This function should work with PEC files produced via adas810 or adas218.

    """
    # find out whether file is metastable resolved
    meta_resolved = path.split('#')[-2][-1]=='r'
    if meta_resolved: print('Identified metastable-resolved PEC file')
    
    with open(path, 'r') as f:
        lines = f.readlines()
    cs = path.split('#')[-1].split('.dat')[0]

    header = lines.pop(0)
    # Get the expected number of lines by reading the header:
    num_lines = int(header.split()[0])
    pec_dict = {}

    for i in range(0, num_lines):
        
        if '----' in lines[0]: 
            _ = lines.pop(0) # separator may exist before each transition

        # Get the wavelength, number of densities and number of temperatures
        # from the first line of the entry:
        l = lines.pop(0)
        header = l.split()

        # sometimes the wavelength and its units are not separated:
        try:
            header = [hh.split('A')[0] for hh in header]
        except:
            # lam and A are separated. Delete 'A' unit.
            header = np.delete(header, 1)

        lam = float(header[0])

        if header[1]=='':
            # 2nd element was empty -- annoyingly, this happens sometimes
            num_den = int(header[2])
            num_temp = int(header[3])
        else:
            num_den = int(header[1])
            num_temp = int(header[2])            

        if meta_resolved:
            # index of metastable state
            INDM = int(header[-3].split('/')[0].split('=')[-1])

        # Get the densities:
        dens = []
        while len(dens) < num_den:
            dens += [float(v) for v in lines.pop(0).split()]
        dens = np.asarray(dens)

        # Get the temperatures:
        temp = []
        while len(temp) < num_temp:
            temp += [float(v) for v in lines.pop(0).split()]
        temp = np.asarray(temp)

        # Get the PEC's:
        PEC = []
        while len(PEC) < num_den:
            PEC.append([])
            while len(PEC[-1]) < num_temp:
                PEC[-1] += [float(v) for v in lines.pop(0).split()]
        PEC = np.asarray(PEC)
        
        # find what kind of rate we are dealing with
        if 'recom' in l.lower(): rate_type = 'recom'
        elif 'excit' in l.lower(): rate_type = 'excit'
        elif 'chexc' in l.lower(): rate_type = 'chexc'
        elif 'drsat' in l.lower(): rate_type = 'drsat'
        else:
            # attempt to report unknown rate type -- this should be fairly robust
            rate_type = l.replace(' ','').lower().split('type=')[1].split('/')[0]

        # create dictionary with keys for each wavelength:
        if lam not in pec_dict:
            pec_dict[lam] = {}                

        # add a key to the pec_dict[lam] dictionary for each type of rate: recom, excit or chexc
        # interpolate PEC on log dens,temp scales
        pec_fun = RectBivariateSpline(
            np.log10(dens),
            np.log10(temp),
            PEC,
            kx=order,
            ky=order
        )
        
        if meta_resolved:
            if rate_type not in pec_dict[lam]:
                pec_dict[lam][rate_type] = {}
            pec_dict[lam][rate_type][f'meta{INDM}'] = pec_fun
        else:
            pec_dict[lam][rate_type] = pec_fun
            
        if lam in plot_lines:

            # use log spacing (important to match well low values)
            ne_eval = 10** np.linspace(np.log10(dens.min()), np.log10(dens.max()), 10)
            Te_eval = 10** np.linspace(np.log10(temp.min()), np.log10(temp.max()), 100)

            NE, TE = np.meshgrid(ne_eval, Te_eval)
            
            PEC_eval = pec_fun.ev(np.log10(NE), np.log10(TE))

            # plot PEC rates
            _ax = _plot_pec(dens,temp,ne_eval,Te_eval,PEC_eval,PEC, lam,cs,rate_type,
                            ax, plot_3d)

            meta_str = ''
            if meta_resolved: meta_str = f' , meta = {INDM}'
            _ax.set_title(cs + r' , $\lambda$ = '+str(lam) +' A, '+rate_type+meta_str)
            plt.tight_layout()

    return pec_dict



def _plot_pec(dens, temp, ne_eval, Te_eval, PEC_eval, PEC, lam,cs,rate_type,
             ax=None, plot_3d=False):
    '''Private method to plot PEC data within :py:func:`~aurora.atomic.read_adf15` function.
    '''

    if ax is None:
        f1 = plt.figure()
        if plot_3d:
            ax1 = f1.add_subplot(1,1,1, projection='3d')
        else:
            ax1 = f1.add_subplot(1,1,1)
    else:
        ax1 = ax

    if plot_3d:
        from mpl_toolkits.mplot3d import Axes3D

        # to plot on linear ne,Te scales:
        #ax1.plot_surface(NE, TE, PEC_eval, alpha=0.5)
        #DENS, TEMP = np.meshgrid(dens, temp)
        
        # log scales don't work in matplotlib 3D, so we plot log of each quantity
        logNE, logTE = np.meshgrid(np.log10(ne_eval), np.log10(Te_eval))
        ax1.plot_surface(logNE, logTE, PEC_eval, alpha=0.5)
        DENS, TEMP = np.meshgrid(np.log10(dens), np.log10(temp))
        
        ax1.scatter(DENS.ravel(), TEMP.ravel(), PEC.T.ravel(), color='b')

        if ax is None:
            ax1.set_xlabel('$log_{10}(n_e)$ [cm$^{-3}$]')
            ax1.set_ylabel('$log_{10}(T_e)$ [eV]')
            ax1.set_zlabel('PEC')

    else:
        # plot in 2D
        labels = ['{:.0e}'.format(ne)+r' $cm^{-3}$' for ne in ne_eval]
        for ine in np.arange(PEC_eval.shape[1]):
            ax1.plot(Te_eval, PEC_eval[:,ine], label=labels[ine])
        ax1.set_xlabel(r'$T_e$ [eV]')
        ax1.set_ylabel('PEC')
        ax1.set_yscale('log')

        ax1.legend(loc='best').set_draggable(True)

    return ax1


def adf04_files():
    '''Collection of trust-worthy ADAS ADF04 files. 
    This function will be moved and expanded in ColRadPy in the near future. 
    '''
    files = {}
    files['Ca'] = {}
    files['Ca']['8'] = 'mglike_lfm14#ca8.dat'
    files['Ca']['9'] = 'nalike_lgy09#ca9.dat'
    files['Ca']['10'] = 'nelike_lgy09#ca10.dat'
    files['Ca']['11'] = 'flike_mcw06#ca11.dat'
    files['Ca']['14'] = 'clike_jm19#ca14.dat'
    files['Ca']['15'] = 'blike_lgy12#ca15.dat'
    files['Ca']['16'] = 'belike_lfm14#ca16.dat'
    files['Ca']['17'] = 'lilike_lgy10#ca17.dat'
    files['Ca']['18'] = 'helike_adw05#ca18.dat'

    # TODO: check quality
    files['Al'] = {}
    files['Al']['11'] = 'helike_adw05#al11.dat'
    files['Al']['10'] = 'lilike_lgy10#al10.dat'
    files['Al']['9'] = 'belike_lfm14#al9.dat'
    files['Al']['8'] = 'blike_lgy12#al8.dat'
    files['Al']['7'] = 'clike_jm19#al7.dat'
    files['Al']['6'] = ''
    files['Al']['5'] = ''
    files['Al']['4'] = 'flike_mcw06#al4.dat'
    files['Al']['3'] = 'nelike_lgy09#al3.dat'
    files['Al']['2'] = 'nalike_lgy09#al2.dat'
    files['Al']['1'] = 'mglike_lfm14#al1.dat'

    # TODO: check quality
    files['F'] = {}
    files['F']['8'] = 'copha#h_hah96f.dat'
    files['F']['7'] = 'helike_adw05#f7.dat'
    files['F']['6'] = 'lilike_lgy10#f6.dat'
    files['F']['5'] = 'belike_lfm14#f5.dat'
    files['F']['4'] = 'blike_lgy12#f4.dat'
    files['F']['3'] = 'clike_jm19#f3.dat'

    return files




def get_colradpy_pec_prof(ion, cs, rhop, ne_cm3, Te_eV, 
                          lam_nm, lam_width_nm, adf04_loc,
                          meta_idxs=[0], pec_threshold=1e-20, pec_units=2, plot=True):
    '''Compute radial profile for Photon Emissivity Coefficients (PEC) for lines within the chosen
    wavelength range using the ColRadPy package. This is an alternative to the option of using 
    the :py:func:`~aurora.radiation.read_adf15` function to read PEC data from an ADAS ADF-15 file and 
    interpolate results on ne,Te grids. 

    Args:
        ion : str
            Ion atomic symbol
        cs : str
            Charge state, given in format like '17', indicating total charge of ion (e.g. '17' would be for Li-like Ca)
        rhop : array (nr,)
            Sqrt of normalized poloidal flux radial array
        ne_cm3 : array (nr,)
            Electron density in :math:`cm^{-3}` units
        Te_eV : array (nr,)
            Electron temperature in eV units
        lam_nm : float
            Center of the wavelength region of interest [nm]
        lam_width_nm : float
            Width of the wavelength region of interest [nm]
        adf04_loc : str
            Location from which ADF04 files listed in :py:func:`adf04_files` should be fetched.

    Keyword Args:
        meta_idxs : list of integers
            List of levels in ADF04 file to be treated as metastable states. Default is [0] (only ground state).
        prec_threshold : float
            Minimum value of PECs to be considered, in :math:`photons \cdot cm^3/s`
        pec_units : int
            If 1, results are given in :math:`photons \cdot cm^3/s`; if 2, they are given in :math:`W.cm^3`. 
            Default is 2.
        plot : bool
            If True, plot lines profiles and total.

    Returns:
        pec_tot_prof : array (nr,)
            Radial profile of PEC intensity, in units of :math:`photons \cdot cm^3/s` (if pec_units=1) or 
            :math:`W \cdot cm^3` (if pec_units=2).
    '''
    try:
        # temporarily import this here, until ColRadPy dependency can be set up properly
        from colradpy import colradpy
    except ImportError:
        raise ValueError('Could not import colradpy. Install this from the Github repo!')
    
    files = adf04_files()

    filepath = adf04_repo+files[ion][cs]
    
    crm = colradpy(filepath, meta_idxs, Te_eV, ne_cm3,temp_dens_pair=True,
                   use_recombination=False,
                   use_recombination_three_body=False)
    
    crm.make_ioniz_from_reduced_ionizrates()
    crm.suppliment_with_ecip()
    crm.make_electron_excitation_rates()
    crm.populate_cr_matrix()   # time consuming step
    crm.solve_quasi_static()

    lams = crm.data['processed']['wave_vac']
    lam_sel_idxs = np.where((lams>lam_nm-lam_width_nm/2.)&(lams<lam_nm+lam_width_nm/2.))[0]
    _lam_sel_nm = lams[lam_sel_idxs]

    pecs = crm.data['processed']['pecs'][lam_sel_idxs,0,:]  # 0 index is for excitation component
    pecs_sel_idxs = np.where((np.max(pecs,axis=1)<pec_threshold))[0]
    pecs_sel = pecs[pecs_sel_idxs,:]
    lam_sel_nm = _lam_sel_nm[pecs_sel_idxs]
    
    # calculate total PEC profile
    pec_tot_prof = np.sum(pecs_sel,axis=0)

    if pec_units==2:
        # convert from photons.cm^3/s to W.cm^3
        mults = constants.h * constants.c / (lam_sel_nm*1e-9)
        pecs_sel *= mults[:,None]
        pec_tot_prof *= mults[:,None]
        
    if plot:
        fig,ax = plt.subplots()
        for ll in np.arange(len(lam_sel_nm)):
            ax.plot(rhop, pecs_sel[ll,:], label=fr'$\lambda={lam_sel_nm[ll]:.5f}$ nm')
        ax.plot(rhop, pec_tot_prof, lw=3.0, c='k', label='Total')
        fig.suptitle(fr'$\lambda={lam_nm} \pm {lam_width_nm/2.}$ nm')
        ax.set_xlabel(r'$\rho_p$')
        ax.set_ylabel('PEC [W cm$^3$]' if phot2energy else 'PEC [ph cm$^3$ s$^{-1}$]')
        ax.legend(loc='best').set_draggable(True)

    return pec_tot_prof



'''
Script demonstrating use of Aurora with explicit radial profiles of impurity neutral sources, 
also allowing users to vary the ne,Te grids based on arbitrary heating, cooling or dilution 
processes. This may be useful, for example, for pellet ablation or massive gas injection studies. 

Run this in IPython, or uncomment plt.show(block=True) at the end. 

jmcclena and sciortino, Nov 2020
'''
import numpy as np
import matplotlib.pyplot as plt
plt.ion()
import omfit_eqdsk, omfit_gapy
import sys, copy, os
from scipy.interpolate import interp1d

# Make sure that package home is added to sys.path
sys.path.append('../')
import aurora


def get_nall(rhop, rhop_grid, ne, nd, nz):
    # This function gives total ion+electron density
    nz_interp = np.zeros([len(rhop),len(nz[0,:])])
    for Z in range(len(nz[0,:])):
        nz_interp[:,Z] = np.interp(rhop, rhop_grid, nz[:,Z])

    nall = ne + nd
    Znz = np.zeros(len(rhop))
    for Z in range(len(nz_interp[0,:])):
        nall += nz_interp[:,Z]
        Znz += Z*nz_interp[:,Z]

    nall += nall + Znz

    return nall, Znz
    
def dilution_cooling(rhop, rhop_grid, ne_old, nd_old, T_old, nz_old, nz_new):
    # This function reduces temperature by the increase in density
    # Assumption of Ti=Te for simplicity

    nall_old, Znz_old  = get_nall(rhop, rhop_grid, ne_old, nd_old, nz_old)
    nall_new, Znz_new = get_nall(rhop, rhop_grid, ne_old, nd_old, nz_new)

    ne_new = ne_old - Znz_old + Znz_new
    T_new = T_old*nall_old/nall_new

    return ne_new, T_new

def radiation_cooling(rhop, rhop_grid, ne, nd, nz, T, Erad):
    # This function subracts radiation power from plasma energy
    nall, ZnZ = get_nall(rhop, rhop_grid, ne, nd, nz)
    Erad = np.interp(rhop, rhop_grid, Erad)
    T_new = T - 2*Erad/(3*nall) # 2/3 to account for stored energy = 3/2 integral p dV
    return T_new

# read in default Aurora namelist
namelist = aurora.default_nml.load_default_namelist()

# Use gfile and statefile in local directory:
examples_dir = os.path.dirname(os.path.abspath(__file__))
geqdsk = omfit_eqdsk.OMFITgeqdsk(examples_dir+'/example.gfile')
inputgacode = omfit_gapy.OMFITgacode(examples_dir+'/example.input.gacode')

# save kinetic profiles on a rhop (sqrt of norm. pol. flux) grid
kp = namelist['kin_profs']
kp['Te']['rhop'] = rhop = kp['ne']['rhop'] = np.sqrt(inputgacode['polflux']/inputgacode['polflux'][-1])
kp['ne']['vals'] = ne_cm3 = inputgacode['ne']*1e13 # 1e19 m^-3 --> cm^-3
kp['Te']['vals'] = Te_eV = inputgacode['Te']*1e3  # keV --> eV
nd_cm3 = copy.deepcopy(ne_cm3)

################## Simulation time steps and duration settings ##################
#
# Update background every n_rep iterations, each of dt [s] length
n_rep = 5
dt = 1e-4

# Total time to run [s] -- will be approximated by nearest multiplier of n_rep*dt
sim_time = 5e-3  # 20 ms
num_sims = int(sim_time/(n_rep*dt))
##################################################################################

# do only a few time steps per "run"
namelist['timing'] = {'dt_increase': np.array([1., 1.   ]),
                      'dt_start': np.array([dt, sim_time]),
                      'steps_per_cycle': np.array([1, 1]),
                      'times': np.array([0. , n_rep*dt])}


# set impurity species and sources rate
imp = namelist['imp'] = 'Ar'

# provide impurity neutral sources on explicit radial and time grids
namelist['explicit_source_time'] = np.linspace(0.,namelist['timing']['times'][-1]*n_rep,99)
namelist['explicit_source_rhop'] = np.linspace(0,1.3,101)
gaussian_rhop = 1e9 * np.exp(- (namelist['explicit_source_rhop']-0.5)**2/(2*0.1**2))
exp_time = np.exp(- namelist['explicit_source_time']/0.02)  # decay over 20ms time scale
namelist['explicit_source_vals'] = gaussian_rhop[None,:]*exp_time[:,None]

fig,ax = plt.subplots(num='Impurity neutral source')
ax.contourf(namelist['explicit_source_rhop'],
             namelist['explicit_source_time'],
             namelist['explicit_source_vals'])
ax.set_xlabel(r'$\rho_p$')
ax.set_ylabel('time [s]')

# Now get aurora setup
asim = aurora.core.aurora_sim(namelist, geqdsk=geqdsk)

# set time-independent transport coefficients (flat D=1 m^2/s, V=-2 cm/s)
D_z = 1e4 * np.ones(len(asim.rvol_grid))  # cm^2/s
V_z = -2e2 * np.ones(len(asim.rvol_grid)) # cm/s

# run Aurora forward model and plot results
out = asim.run_aurora(D_z, V_z, plot=False)
nz_all = out[0] # impurity charge state densities are the first element of "out"
nz_init = nz_all[:,:,-1]

# calculate dilution cooling
rad = aurora.compute_rad(
    imp, out[0].transpose(2,1,0), asim.ne, asim.Te, prad_flag=True)
tot_rad_dens = rad['tot'] # W/cm^3
line_rad_all = rad['line_rad'].T # W/cm^3
time_grid = copy.deepcopy(asim.time_grid)

# modify background temperature and density profiles based on tot_rad_dens
rhop_grid = asim.rhop_grid
Erad = np.trapz(tot_rad_dens[-1*n_rep:,:],axis=0,dx=dt) * 1.6e-13
Te_eV = radiation_cooling(rhop, rhop_grid, ne_cm3, nd_cm3, nz_init, Te_eV, Erad)
ne_cm3, Te_eV = dilution_cooling(rhop, rhop_grid, ne_cm3, nd_cm3, Te_eV, nz_init*0., nz_init)
kp['Te']['vals'] = Te_eV
kp['ne']['vals'] = ne_cm3

# update kinetic profile dependencies:
asim.setup_kin_profs_depts()

Te_all = []
Te_all.append(Te_eV)
for i in np.arange(num_sims):
    # Update time array
    namelist['timing']['times'] = np.array([i*n_rep*dt+dt, (i+1)*n_rep*dt])
    asim.setup_grids()
    
    # get charge state densities from latest time step
    nz_old = nz_all[:,:,-1*n_rep]
    nz_init = nz_all[:,:,-1]
    out = asim.run_aurora(D_z, V_z, nz_init=nz_init, plot=False)
    nz_all = np.dstack((nz_all, out[0]))

    rad = aurora.compute_rad(
        imp, out[0].transpose(2,1,0), asim.ne, asim.Te, prad_flag=True)
    tot_rad_dens = rad['tot'] # W/cm^3
    line_rad_all = np.dstack((line_rad_all, rad['line_rad'].T))
    time_grid = np.concatenate((time_grid, asim.time_grid))
    
    # modify background temperature and density profiles
    rhop_grid = asim.rhop_grid
    Erad = np.trapz(tot_rad_dens[-1*n_rep:,:], axis=0) * 1.6e-13
    Te_eV = radiation_cooling(rhop, rhop_grid, ne_cm3, nd_cm3, nz_init, Te_eV, Erad)
    ne_cm3, Te_eV = dilution_cooling(rhop, rhop_grid, ne_cm3, nd_cm3, Te_eV, nz_old, nz_init)
    kp['Te']['vals'] = Te_eV
    kp['ne']['vals'] = ne_cm3
    Te_all.append(Te_eV)
    # update kinetic profile dependencies:
    asim.setup_kin_profs_depts()
    

# plot charge state distributions over radius and time
aurora.plot_tools.slider_plot(asim.rvol_grid, time_grid, nz_all.transpose(1,0,2),
                              xlabel=r'$r_V$ [cm]', ylabel='time [s]', zlabel=r'$n_z$ [$cm^{-3}$]',
                              labels=[str(i) for i in np.arange(0,nz_all.shape[1])],
                              plot_sum=True, x_line=asim.rvol_lcfs)

# plot radiation profiles over radius and time
aurora.slider_plot(asim.rvol_grid, time_grid, line_rad_all.transpose(1,0,2),
                              xlabel=r'$r_V$ [cm]', ylabel='time [s]', zlabel=r'Line radiation [$MW/m^3$]',
                              labels=[str(i) for i in np.arange(0,nz_all.shape[1])],
                              plot_sum=True, x_line=asim.rvol_lcfs)


_Te_all = np.array(Te_all).T
Te_arr = np.reshape(_Te_all, (1,len(rhop),len(time_grid[::5])))
aurora.slider_plot(rhop, time_grid[::5], Te_arr, xlabel=r'$rho_p$', ylabel='time [s]', zlabel=r'Te [eV]')

#plt.show(block=True)




'''Aurora tools to read SOLPS results and extract atomic neutral density from EIRENE output.
These enable examination of charge exchange recombination and radiation due to the interaction 
of heavy ions and thermal neutrals. 

F.Sciortino & R.Reksoatmodjo, 2021
'''
import matplotlib.pyplot as plt
import os
import numpy as np
plt.ion()
from scipy.interpolate import griddata
from matplotlib import cm
import matplotlib.tri as tri
from heapq import nsmallest
import warnings

from . import plot_tools
from . import coords

class solps_case:
    def __init__(self, path, geqdsk, solps_run='',
                 case_num=0, form='extracted', extracted_labels=None):
        '''Read SOLPS output and prepare for Aurora impurity-neutral analysis. 

        Parameters
        -----------------
        path : str
            Path to output files. If these are "extracted" files from SOLPS (form='extracted'),
            then this is the path to the disk location where each of the required files
            can be found. 
            If form='full', this path indicates where to find the directory named "baserun"
            and the 'solps_run' one.
        geqdsk : str or `omfit_geqdsk` class instance
            Path to the geqdsk to load from disk, or instance of the `omfit_geqdsk` class that
            contains the processed gEQDSK file already. 
        solps_run : str
            If form='full', this string specifies the directory (relative to the given path)
            where case-specific files for a SOLPS run can be found (e.g. 'b2fstate').
        case_num : int
            Index/integer identifying the SOLPS case of interest. 
        form : str
            Form of SOLPS output to be loaded, one of {'full','extracted'}. 
            The 'full' output consists of 'b2fstate', 'b2fgmtry', 'fort.44', etc.
            The 'extracted' output consists of individual files containing one quantity each.
            If form='extracted', the 'extracted_labels' argument allows to specify files 
            nomanclature. 
        extracted_labels : dict
            Only used if form='extracted', this dictionary allows specification of the names of
            files/variables to be extracted. Default is to use
            [RadLoc, VertLoc, Ne, Te, NeuDen, NeuTemp, MolDen, MolTemp, Ti]
            each of which will be expected to have "case_num" appended to the name of the file.
        '''

        self.path = path
        self.solps_run = solps_run
        self.case_num = case_num
        self.form = form
        
        if isinstance(geqdsk, str): # user passed a path to a gfile on disk
            import omfit_eqdsk # import omfit_eqdsk here to avoid issues with docs and packaging
            self.geqdsk = omfit_eqdsk.OMFITgeqdsk(geqdsk)
        else:
            self.geqdsk = geqdsk

        self.labels = {'ne':r'log($n_e$) [$m^{-3}$]' , 'Te':r'log($T_e$) [eV]',
                      'nn':'log($n_n$) [$m^{-3}$]' , 'Tn':'log($T_n$) [eV]',
                      'nm':'log($n_m$) [$m^{-3}$]' , 'Tm':'log($T_m$) [eV]',
                      'Ti':r'log($T_i$) [eV]',
            }
            
        if form=='extracted':
            self.ext_names_map = extracted_labels
            if self.ext_names_map is None: # set default map to identify extracted data file names
                self.ext_names_map = {'ne':'Ne','Te':'Te','nn':'NeuDen','Tn':'NeuTemp',
                                      'nm':'MolDen','Tm':'MolTemp','Ti':'Ti',
                                      'R':'RadLoc','Z':'VertLoc'}

            tmp = np.loadtxt(path+f'/{self.ext_names_map["R"]}{case_num}')

            self.ny = int(np.max(tmp[:,1]))
            self.nx = int(np.max(tmp[:,0]))

            self.unit_p = self.nx//4
            self.unit_r = self.ny//2

            self.double_null = False # TODO: generalize for double null shapes!

        elif form=='full':
            import omfit_solps # import omfit_solps here to avoid issues with docs and packaging
            self.b2fstate = omfit_solps.OMFITsolps(path+os.sep+solps_run+os.sep+'b2fstate')
            self.geom = omfit_solps.OMFITsolps(path+os.sep+'baserun'+os.sep+'b2fgmtry')

            self.nx,self.ny = self.geom['nx,ny']
            
            # (0,:,:): lower left corner, (1,:,:): lower right corner
            # (2,:,:): upper left corner, (3,:,:): upper right corner.
            
            self.crx = self.geom['crx'].reshape(4,self.ny+2,self.nx+2)  # horizontal 
            self.cry = self.geom['cry'].reshape(4,self.ny+2,self.nx+2)  # vertical

            # figure out if single or double null
            self.double_null = self.cry.shape[2]%4==0
            
            # uncertain units for splitting of radial and poloidal regions...
            self.unit_p = (self.crx.shape[2]-4 if self.double_null else self.crx.shape[2]-2)//4
            self.unit_r = (self.crx.shape[1]-2 if self.double_null else self.crx.shape[1])//2 

        if self.double_null:
            # TODO: unlikely to be correct, double null generalization still incomplete!
            
            # Obtain indices for chosen radial regions
            _R_idxs = np.array([],dtype=int)
            _R_idxs = np.concatenate((_R_idxs, np.arange(self.unit_r)))  # PFR and core
            self.R_idxs = np.concatenate((_R_idxs, np.arange(self.unit_r,2*self.unit_r))) # open-SOL
            
            # obtain indices for chosen poloidal regions
            _P_idxs = np.array([],dtype=int)
            _P_idxs = np.concatenate((_P_idxs, np.arange(self.unit_p+1)))  # Inner PFR
            _P_idxs = np.concatenate((_P_idxs, np.arange(self.unit_p+1,3*self.unit_p+1)))  # core/open SOL
            self.P_idxs = np.concatenate((_P_idxs, np.arange(3*self.unit_p+1,4*self.unit_p+2)))  # outer PFR
            
        else:  # upper or lower single null
            # Obtain indices for chosen radial regions
            _R_idxs = np.array([],dtype=int)
            _R_idxs = np.concatenate((_R_idxs, np.arange(self.unit_r+1)))  # PFR and core
            self.R_idxs = np.concatenate((_R_idxs, np.arange(self.unit_r+1,2*self.unit_r+2))) # open-SOL
            
            # obtain indices for chosen poloidal regions
            _P_idxs = np.array([],dtype=int)
            _P_idxs = np.concatenate((_P_idxs, np.arange(self.unit_p+1)))  # Inner PFR
            _P_idxs = np.concatenate((_P_idxs, np.arange(self.unit_p+1,3*self.unit_p+1)))  # core/open SOL
            self.P_idxs = np.concatenate((_P_idxs, np.arange(3*self.unit_p+1,4*self.unit_p+2)))  # outer PFR



    def load_data(self, fields=None, P_idxs=None, R_idxs=None,
                  Rmin=None, Rmax=None, Pmin=None, Pmax=None):
        '''Load SOLPS output for each of the needed quantities ('extracted' form)

        Parameters
        -----------------
        fields : list or array
            List of fields to fetch from SOLPS output. If left to None, by default uses
            self.labels.keys(), i.e. ['ne','Te','nn','nT','nm','Tm','Ti']
        P_idxs : list or array
            Poloidal indices to load.
        R_idxs : list or array
            Radial indices to load.
        Rmin : int or None.
            Minimum major radius index to load, if R_idxs is not given
        Rmax : int or None
            Maximum major radius index to load, if R_idxs is not given
        Pmin : int or None
            Minimum poloidal index to load, if P_idxs is not given
        Pmax : int or None
            Maximum poloidal index to load, if P_idxs is not given
        
        Returns
        ------------
        quants : dict
            Dictionary containing 'R','Z' coordinates for 2D maps of each field requested by user.
        '''
        if P_idxs is None:
            if Pmax is None: Pmax = self.ny
            if Pmin is None: Pmin = 0
            P_idxs = np.arange(Pmin,Pmax)
        else:
            pass # user provided list of poloidal grid indices

        if R_idxs is None:
            if Rmax is None: Rmax = self.nx
            if Rmin is None: Rmin = 0
            R_idxs = np.arange(Rmin,Rmax)
        else:
            pass # user provided list of radial grid indices

        if fields is None:
            fields = list(self.labels.keys())
            
        self.quants = quants = {}
        
        if self.form=='extracted':

            self.R = np.atleast_2d(
                np.loadtxt(self.path+f'/{self.ext_names_map["R"]}{self.case_num}',
                           usecols=(3)).reshape((self.ny+2,self.nx+2))[:,P_idxs])[R_idxs,:]
            self.Z = np.atleast_2d(
                np.loadtxt(self.path+f'/{self.ext_names_map["Z"]}{self.case_num}',
                           usecols=(3)).reshape((self.ny+2,self.nx+2))[:,P_idxs])[R_idxs,:]

            for field in fields:
                tmp =  np.loadtxt(self.path+f'/{self.ext_names_map[field]}{self.case_num}',
                                  usecols=(3)).reshape((self.ny+2,self.nx+2))
                quants[field] = np.atleast_2d(tmp[:,P_idxs])[R_idxs,:]

            self.triang_b2 = tri.Triangulation(self.R.flatten(), self.Z.flatten())
            self.triang_eirene = self.triang_b2 # triangulation for EIRENE is mapped to b2 grid
    
        elif self.form=='full':

            # eliminate end (buffer) points of grid
            self.R = np.mean(self.crx,axis=0)[R_idxs,:][:,P_idxs]
            self.Z = np.mean(self.cry,axis=0)[R_idxs,:][:,P_idxs]

            for field in fields:
                if field in ['ne','Te','Ti']:
                    quants[field] = self.b2fstate[field.lower()][R_idxs,:][:,P_idxs]

            self.eirene_out = self.load_eirene_output(files=['fort.44'])  #'fort.46'

            ########
            ## Read on EIRENE mesh
            # molecular density and temperature
            # quants['nm'] = self.eirene_out['fort.46']['pdenm']
            # quants['Tm']= self.eirene_out['fort.46']['edenm']

            # # group atomic neutral density from all H isotopes
            # spmask = self.b2fstate['zn']==1
            # quants['nn'] = np.sum(
            #     self.eirene_out['fort.46']['pdena'].reshape(-1,sum(spmask)),
            #     axis=1)

            # quants['Tn'] = np.sum(
            #     self.eirene_out['fort.46']['edena'].reshape(-1,sum(spmask)),
            #     axis=1)
            ####

            tmp = self.eirene_out['fort.44']['dmb2'].reshape((self.ny,self.nx))
            quants['nm'] = np.atleast_2d(tmp)[R_idxs,:][:,P_idxs]
            tmp = self.eirene_out['fort.44']['tmb2'].reshape((self.ny,self.nx))
            quants['Tm'] = np.atleast_2d(tmp)[R_idxs,:][:,P_idxs]

            # group atomic neutral density from all H isotopes
            spmask = self.b2fstate['zn']==1
            tmp = np.sum(
                self.eirene_out['fort.44']['dab2'].reshape(-1,sum(spmask)),
                axis=1).reshape((self.ny,self.nx))
            quants['nn'] = np.atleast_2d(tmp)[R_idxs,:][:,P_idxs]
            tmp = np.sum(
                self.eirene_out['fort.44']['tab2'].reshape(-1,sum(spmask)),
                axis=1).reshape((self.ny,self.nx))
            quants['Tn'] = np.atleast_2d(tmp)[R_idxs,:][:,P_idxs]
            
            self.triang_b2 = tri.Triangulation(self.R.flatten(), self.Z.flatten())
            self.triang_eirene = self.eirene_out['triang']
            
            
            


    def load_eirene_output(self, files = ['fort.44','fort.46']):
        '''Load result from one of the fort.* files with EIRENE output, 
        as indicated by the "files" list argument.

        Parameters
        -----------------
        files : list or array-like
            EIRENE output files to read. Default is to load all files for which this method has been tested. 

        Returns
        ------------
        eirene_out : dict
            Dictionary for each loaded file containing a subdictionary with keys for each loaded field from each file. 
        '''
        eirene_out = {}
        
        for filename in files: 
            eirene_out[filename] = {}
            # load each of these files into dictionary structures
            with open(self.path +os.sep+self.solps_run+os.sep+filename, 'r') as f:
                contents = f.readlines()
            ii=6

            while ii<len(contents[ii:]):
                if  contents[ii].startswith('*eirene'):
                    key = contents[ii].split()[3]
                    eirene_out[filename][key] = []
                else:
                    [eirene_out[filename][key].append(float(val)) for val in contents[ii].strip().split()]
                ii+=1

            for key in eirene_out[filename]:
                eirene_out[filename][key] = np.array(eirene_out[filename][key])

        # Now load fort.33 file with EIRENE nodes and cells
        Nodes=np.fromfile(self.path+os.sep+'baserun'+os.sep+'fort.33',sep=' ')
        NN=int(Nodes[0])
        XNodes=Nodes[1:NN+1]/100  # cm -->m
        YNodes=Nodes[NN+1:]/100
        
        # EIRENE triangulation
        Triangles = np.loadtxt(
            self.path+os.sep+'baserun'+os.sep+'fort.34',skiprows=1, usecols=(1,2,3))

        eirene_out['triang'] =tri.Triangulation(XNodes,YNodes,triangles=(Triangles-1))

        return eirene_out


    def process_solps_data(self, fields=None, P_idxs=None, R_idxs=None, plot=False):
        '''Load and process SOLPS to permit clear plotting. 
        
        Parameters
        -----------------
        fields : dict
            Dictionary containing SOLPS outputs to process. Keys indicate the quantity, value its label
            (only used for plotting). If left to None, defaults fields of 'ne','Te','nn' and 'Tn' are used.
        P_idxs : list or array
            Poloidal indices to load.
        R_idxs : list or array
            Radial indices to load.
        plot : bool
            If True, plot results for all loaded 2D quantities. 

        Returns
        ------------
        quants : dict
            Dictionary containing 'R','Z' coordinates for 2D maps of each field requested by user.
            Quantities are processed and masked to facilitate plotting.
        '''
        if fields is None:
            fields = ['ne','Te','nn','Tn'] #list(self.labels.keys())

        if P_idxs is None:
            P_idxs = self.P_idxs
        if R_idxs is None:
            R_idxs = self.R_idxs

        # load data arrays
        self.load_data(fields = fields, P_idxs=P_idxs, R_idxs=R_idxs)

        # apply masking based on maximum side-length
        self.max_mask_len = 10* (np.max(self.Z)-np.min(self.Z))/(self.nx) #ROT to get an appropriate edge masking length
        self.triang_b2 = apply_mask(self.triang_b2, self.geqdsk, max_mask_len=self.max_mask_len)

        # set zero densities equal to min to avoid log issues
        self.quants['nn'][self.quants['nn']==0.0] = nsmallest(2,np.unique(self.quants['nn'].flatten()))[1]

        if plot:

            fig,axs = plt.subplots(2,2, figsize=(9,12),sharex=True) 
            ax = axs.flatten()

            cbars = [];
            for ii,field in enumerate(fields):
                label = fields[field]

                ax[ii].plot(self.geqdsk['RBBBS'],self.geqdsk['ZBBBS'],c='k')
                cntr = ax[ii].tricontourf(self.triang_b2, np.log10(self.quants[field]).flatten(),
                                           cmap=cm.magma,levels=300)

                # create draggable colorbar
                cbars.append( plt.colorbar(cntr, format='%.3g', ax=ax[ii]))
                cbars[-1].ax.set_title(label)
                cbars[-1] = plot_tools.DraggableColorbar(cbars[-1],cntr)
                cbars[-1].connect()

                ax[ii].axis('equal')
                ax[ii].set_xlabel('R [m]')
                ax[ii].set_ylabel('Z [m]')



    def plot_2d_vals(self, vals, label='', use_triang=True, max_mask_len=None):
        '''Method to plot 2D fields over the R and Z grids. 
        Colorbars are set to be manually adjustable, allowing variable image saturation.

        Parameters
        -----------------
        vals : array, (ny,nx)
            2D array containing a variable of interest, on the same grid as the 
            R and Z attributes. 

        Returns
        ------------
        label : str
            Label describing the quantity being plotted.
        use_triang : bool
            If True, use a triangulation to plot results on B2 grids, attempting
            to mask unwanted spurious edges. Alternatively, use a simple filled 
            contour plot, which however may show unphysical features at the edges
            of the B2 grid. Default is True (use the triangulation).  
        max_mask_len : float
            Optional maximum length of triangulation segments to be masked. If left to 
            None, the default rule-of-thumb length for this variable will be used, but 
            this may need hand adjustments for specific use cases.
        '''

        fig,ax0 = plt.subplots(figsize=(8,11))
        ax0.axis('equal')
        ax0.set_xlabel('R [m]'); ax0.set_ylabel('Z [m]')

        # plot LCFS
        ax0.plot(self.geqdsk['RBBBS'],self.geqdsk['ZBBBS'],c='k')
        
        if use_triang:
            # possibly modify edge masking of triangulation
            if max_mask_len is not None:
                self.max_mask_len = float(max_mask_len)
                self.triang_b2 = apply_mask(self.triang_b2, self.geqdsk, max_mask_len=max_mask_len)
            
            # triangulation from b2 R and Z grid
            cntr = ax0.tricontourf(self.triang_b2, vals.flatten(), cmap=cm.magma,levels=300)
        else:
            # Simpler plotting, with no triangulation
            cntr = ax0.contourf(self.R, self.Z, vals)

        cbar = plt.colorbar(cntr, format='%.3g', ax=ax0)
        cbar.set_label(label)
        cbar = plot_tools.DraggableColorbar(cbar,cntr)
        cbar.connect()


    def get_radial_prof(self, quant='nn', dz_mm=5, plot=False):
        '''Extract radial profiles of a quantity "quant" from the SOLPS run. 
        This function returns profiles on the low- (LFS) and high-field-side (HFS) midplane, 
        as well as flux surface averaged (FSA) ones. 

        Parameters
        -----------------
        quant : str
            Quantity of interest. Default is 'nn' (neutral atomic density). See self.labels.keys()
            for other options.
        dz_mm : float
            Vertical range [mm] over which quantity should be averaged near the midplane. 
            Mean and standard deviation of profiles on the LFS and HFS will be returned based on
            variations of atomic neutral density within this vertical span.
            Note that this does not apply to the FSA calculation. Default is 5 mm.
        plot : bool
            If True, plot radial profiles. 

        Returns
        ------------
        rhop_fsa : 1D array
            Sqrt of poloidal flux grid on which FSA profiles are given.
        prof_fsa : 1D array
            FSA profile on rhop_fsa grid.
        rhop_LFS : 1D array
            Sqrt of poloidal flux grid on which LFS profile  (prof_LFS) is given.
        prof_LFS : 1D array
            Mean LFS midpane profile on rhop_LFS grid.
        prof_LFS_std : 1D array
            Standard deviation of LFS midplane profile on the rhop_LFS grid, based on variations 
            within +/-`dz_mm`/2 millimeters from the midplane. 
        rhop_HFS : 1D array
            Sqrt of poloidal flux grid on which the midplane HFS profile (prof_HFS) is given.
        prof_HFS : 1D array
            Mean HFS midplane profile on rhop_HFS grid.
        prof_HFS_std : 1D array
            Standard deviation of HFS midplane profile on rhop_HFS grid, based on variations 
            within +/-`dz_mm`/2 millimeters from the midplane.
        '''
        
        rhop_2D = coords.get_rhop_RZ(self.R,self.Z, self.geqdsk)
        
        # evaluate FSA radial profile inside the LCFS
        def avg_function(r,z):
            if any(coords.get_rhop_RZ(r,z, self.geqdsk)<np.min(rhop_2D)):
                return np.nan
            else:
                return griddata((self.R.flatten(),self.Z.flatten()), self.quants[quant].flatten(),
                                (r,z), method='linear')

        prof_FSA = self.geqdsk['fluxSurfaces'].surfAvg(function=avg_function)
        rhop_FSA = np.sqrt(self.geqdsk['fluxSurfaces']['geo']['psin'])

        # get R axes on the midplane on the LFS and HFS
        _dz = (np.max(self.Z) - np.min(self.Z))/((self.nx+self.ny)/10.) # rule-of-thumb to identify vertical resolution
        mask = (self.Z.flatten()>-_dz)&(self.Z.flatten()<_dz)
        R_midplane = self.R.flatten()[mask]
        
        R_midplane_lfs = R_midplane[R_midplane>self.geqdsk['RMAXIS']]
        _R_LFS = np.linspace(np.min(R_midplane_lfs), np.max(R_midplane_lfs),1000)
        
        R_midplane_hfs = R_midplane[R_midplane<self.geqdsk['RMAXIS']]
        _R_HFS = np.linspace(np.min(R_midplane_hfs), np.max(R_midplane_hfs),1000)

        # get midplane radial profile...
        # ...on the LFS:
        _prof_LFS = griddata((self.R.flatten(),self.Z.flatten()), self.quants[quant].flatten(),
                             (_R_LFS,0.5*dz_mm*1e-3*np.random.random(len(_R_LFS))),
                             #(_R_LFS,np.zeros_like(_R_LFS)),
                             method='linear')
        _prof_LFS[_prof_LFS<0]=np.nan
        R_LFS = np.linspace(np.min(R_midplane_lfs), np.max(R_midplane_lfs),100)
        rhop_LFS = coords.get_rhop_RZ(R_LFS,np.zeros_like(R_LFS), self.geqdsk)

        # ... and on the HFS:
        _prof_HFS = griddata((self.R.flatten(),self.Z.flatten()), self.quants[quant].flatten(),
                             #(_R_HFS, np.zeros_like(_R_HFS)),
                             (_R_HFS,0.5*dz_mm*1e-3*np.random.random(len(_R_HFS))),
                             method='linear')
        _prof_HFS[_prof_HFS<0]=np.nan
        R_HFS = np.linspace(np.min(R_midplane_hfs), np.max(R_midplane_hfs),100)   
        rhop_HFS = coords.get_rhop_RZ(R_HFS,np.zeros_like(R_HFS), self.geqdsk)
        
        with warnings.catch_warnings():
            warnings.filterwarnings('ignore') # we might take the mean of slices with only nan's, but that's OK
            prof_LFS = np.nanmean(_prof_LFS.reshape(-1,10),axis=1) # average across 10 near points
            prof_HFS = np.nanmean(_prof_HFS.reshape(-1,10),axis=1)  # average across 10 near points

        # now obtain also the simple poloidal grid slice near the midplane (LFS and HFS)
        # These are commonly used for SOLPS analysis, using the JXA and JXI indices (which we re-compute here)
        Z_core = self.Z[self.unit_r:2*self.unit_r,self.unit_p:3*self.unit_p]
        R_core = self.R[self.unit_r:2*self.unit_r,self.unit_p:3*self.unit_p]

        # find indeces of poloidal grid nearest to Z=0 in the innermost radial shell
        midplane_LFS_idx = np.argmin(np.abs(Z_core[0,R_core[0,:]>self.geqdsk['RMAXIS']]))
        midplane_HFS_idx = np.argmin(np.abs(Z_core[0,R_core[0,:]<self.geqdsk['RMAXIS']]))

        # convert to indices on self.Z and self.R
        JXI = self.unit_p + np.arange(Z_core.shape[1])[R_core[0,:]<self.geqdsk['RMAXIS']][midplane_HFS_idx]  # HFS_mid_pol_idx
        JXA = self.unit_p + np.arange(Z_core.shape[1])[R_core[0,:]>self.geqdsk['RMAXIS']][midplane_LFS_idx] # LFS_mid_pol_idx

        # find rhop along midplane grid chords
        rhop_chord_HFS = coords.get_rhop_RZ(self.R[:,JXI],self.Z[:,JXI], self.geqdsk)
        rhop_chord_LFS = coords.get_rhop_RZ(self.R[:,JXA],self.Z[:,JXA], self.geqdsk)

        if plot:
            # plot quantties on linear scale -- need to adapt labels
            lab = fr'${self.labels[quant].split("$")[1]}$ [${self.labels[quant].split("$")[3]}$]'
            
            # compare FSA radial profiles with midplane (LFS and HFS) ones
            fig,ax = plt.subplots()
            ax.plot(rhop_FSA, prof_FSA, label='FSA')
            ax.plot(rhop_LFS, prof_LFS, label='LFS midplane')
            ax.plot(rhop_HFS, prof_HFS, label='HFS midplane')
            ax.plot(rhop_chord_LFS, self.quants[quant][:,JXA], label='LFS grid midplane')
            ax.plot(rhop_chord_HFS, self.quants[quant][:,JXI], label='HFS grid midplane')
            ax.set_xlabel(r'$\rho_p$')
            ax.set_ylabel(lab)
            ax.legend(loc='best').set_draggable(True)
            plt.tight_layout()

        return rhop_FSA, prof_FSA, rhop_LFS, prof_LFS, rhop_HFS, prof_HFS


def apply_mask(triang, geqdsk, max_mask_len=0.4, mask_up=False, mask_down=False):
    '''Function to apply basic masking to a matplolib triangulation. This type of masking
    is useful to avoid having triangulation edges going outside of the true simulation
    grid. 

    Parameters
    -----------------
    triang : instance of matplotlib.tri.triangulation.Triangulation
        Matplotlib triangulation object for the (R,Z) grid. 
    geqdsk : dict
        Dictionary containing gEQDSK file values as processed by the `omfit_eqdsk`
        package. 
    max_mask_len : float
        Maximum length [m] of segments within the triangulation. Segments longer
        than this value will not be plotted. This helps avoiding triangulation 
        over regions where no data should be plotted, beyond the actual simulation
        grid.
    mask_up : bool
        If True, values in the upper vertical half of the mesh are masked. 
        Default is False.
    mask_down : bool
        If True, values in the lower vertical half of the mesh are masked. 
        Default is False.
      
    Returns
    ------------
    triang : instance of matplotlib.tri.triangulation.Triangulation
        Masked instance of the input matplotlib triangulation object.
    '''
    triangles = triang.triangles
    x = triang.x; y = triang.y
    
    # Find triangles with sides longer than max_mask_len
    xtri = x[triangles] - np.roll(x[triangles], 1, axis=1)
    ytri = y[triangles] - np.roll(y[triangles], 1, axis=1)
    maxi = np.max(np.sqrt(xtri**2 + ytri**2), axis=1)
    cond_maxd = maxi > max_mask_len

    # second condition: mask upper and/or lower part of the grid
    if mask_up:
        cond_up = np.mean(y[triangles],axis=1)>0.0
    else:
        cond_up = np.mean(y[triangles],axis=1)<1e10  # all True
    if mask_down:
        cond_down = np.mean(y[triangles],axis=1)<=0.0
    else:
        cond_down = np.mean(y[triangles],axis=1)<1e10  # all True

    cond_y = cond_up & cond_down

    # apply mask to points within the LCFS (no triangulation issues here)
    rhop_triangs = coords.get_rhop_RZ(np.mean(x[triangles],axis=1),
                             np.mean(y[triangles],axis=1),
                             geqdsk) 
    center_mask = rhop_triangs < np.min(coords.get_rhop_RZ(x,y,geqdsk))

    # apply masking
    triang.set_mask((cond_maxd & cond_y) | center_mask)

    return triang



'''
Script to test functionality from namelist creation to run and postprocessing.

It is recommended to run this in IPython.
'''

import numpy as np
import matplotlib.pyplot as plt
plt.ion()
import omfit_eqdsk, omfit_gapy
import sys, os
from scipy.interpolate import interp1d

# Make sure that package home is added to sys.path
sys.path.append('../')
import aurora

try: # pass any argument via the command line to show plots
    plot = len(sys.argv)>1
except:
    plot = False

# read in default Aurora namelist
namelist = aurora.default_nml.load_default_namelist()

# Use gfile and statefile in local directory:
examples_dir = os.path.dirname(os.path.abspath(__file__))
geqdsk = omfit_eqdsk.OMFITgeqdsk(examples_dir+'/example.gfile')
inputgacode = omfit_gapy.OMFITgacode(examples_dir+'/example.input.gacode')

# save kinetic profiles on a rhop (sqrt of norm. pol. flux) grid
kp = namelist['kin_profs']
kp['Te']['rhop'] = kp['ne']['rhop'] = np.sqrt(inputgacode['polflux']/inputgacode['polflux'][-1])
kp['ne']['vals'] = inputgacode['ne']*1e13 # 1e19 m^-3 --> cm^-3
kp['Te']['vals'] = inputgacode['Te']*1e3  # keV --> eV

# set impurity species and sources rate
imp = namelist['imp'] = 'Ar'
namelist['source_type'] = 'const'
namelist['Phi0'] = 2e20  # particles/s

# Change radial resolution from default:
#namelist['dr_0']=0.2
#namelist['dr_1']=0.02

# Change time resolution from default:
#namelist['timing']['dt_increase'] = np.array([1.001, 1.0, 1.])
#namelist['timing']['dt_start'] = np.array([1e-5, 5e-5, 0.001])
#namelist['timing']['steps_per_cycle'] = np.array([1,1,1])
#namelist['timing']['times'] = np.array([0.,0.05,0.2])

# Now get aurora setup
asim = aurora.core.aurora_sim(namelist, geqdsk=geqdsk)

# check radial grid:
_ = aurora.create_radial_grid(namelist,plot=plot)

# check time grid:
_ = aurora.create_time_grid(namelist['timing'], plot=plot)

# set time-independent transport coefficients (flat D=1 m^2/s, V=-2 cm/s)
D_z = 1e4 * np.ones(len(asim.rvol_grid))  # cm^2/s
V_z = -2e2 * np.ones(len(asim.rvol_grid)) # cm/s

# run Aurora forward model and plot results
out = asim.run_aurora(D_z, V_z, plot=plot)

# extract densities and particle numbers in each simulation reservoir
nz, N_wall, N_div, N_pump, N_ret, N_tsu, N_dsu, N_dsul, rcld_rate, rclw_rate = out

# add radiation
asim.rad = aurora.compute_rad(imp, nz.transpose(2,1,0), asim.ne, asim.Te,
                              prad_flag=True, thermal_cx_rad_flag=False, 
                              spectral_brem_flag=False, sxr_flag=False)

if plot:
    # plot radiation profiles over radius and time
    aurora.slider_plot(asim.rvol_grid, asim.time_out, asim.rad['line_rad'].transpose(1,2,0),
                       xlabel=r'$r_V$ [cm]', ylabel='time [s]', zlabel=r'Line radiation [$MW/m^3$]',
                       labels=[str(i) for i in np.arange(0,nz.shape[1])],
                       plot_sum=True, x_line=asim.rvol_lcfs)


# plot Delta-Zeff profiles over radius and time
asim.calc_Zeff()

if plot:
    # plot variation of Zeff due to simulated impurity:
    aurora.slider_plot(asim.rvol_grid, asim.time_out, asim.delta_Zeff.transpose(1,0,2),
                       xlabel=r'$r_V$ [cm]', ylabel='time [s]', zlabel=r'$\Delta$ $Z_{eff}$',
                       labels=[str(i) for i in np.arange(0,nz.shape[1])],
                       plot_sum=True,x_line=asim.rvol_lcfs)
    

# plot expected centrifugal asymmetry from finite rotation
rhop_gacode = aurora.rad_coord_transform(inputgacode['rho'],'rhon','rhop', asim.geqdsk)

# omega appears unreliable near axis in input.gacode
omega = interp1d(rhop_gacode[3:], inputgacode['omega0'][3:],
                 bounds_error=False,fill_value='extrapolate')(asim.rhop_grid)

# obtain net Zeff in this discharge (exclude last point, unreliable)
Zeff = interp1d(rhop_gacode[:-1], inputgacode['z_eff'][:-1],
                 bounds_error=False,fill_value='extrapolate')(asim.rhop_grid)

# Obtain estimates for centrifigal asymmetry and plot expected 2D distribution inside LCFS
asim.centrifugal_asym(omega, Zeff, plot=plot)



'''
Script to test quality of flux-surface-averaging (FSA) procedure by evaluating the normalized ionization rate, 
or equivalently by comparing the ionization rate of impurities with the parallel transport transit rate. 
This is inspired by Dux et al. NF 2020.

It is recommended to run this in IPython.
'''
import numpy as np
import matplotlib.pyplot as plt
plt.ion()
import omfit_eqdsk, omfit_gapy
import scipy,sys,os
import time
from scipy.interpolate import interp1d

# Make sure that package home is added to sys.path
sys.path.append('../')
import aurora

# read in default Aurora namelist
namelist = aurora.default_nml.load_default_namelist()
kp = namelist['kin_profs']

# Use gfile and statefile in local directory:
examples_dir = os.path.dirname(os.path.abspath(__file__))
geqdsk = omfit_eqdsk.OMFITgeqdsk(examples_dir+'/example.gfile')
inputgacode = omfit_gapy.OMFITgacode(examples_dir+'/example.input.gacode')

# transform rho_phi (=sqrt toroidal flux) into rho_psi (=sqrt poloidal flux) and save kinetic profiles
rhop = kp['Te']['rhop'] = kp['ne']['rhop'] = np.sqrt(inputgacode['polflux']/inputgacode['polflux'][-1])
kp['ne']['vals'] = inputgacode['ne'][None,:]*1e13 # 1e19 m^-3 --> cm^-3
kp['Te']['vals'] = inputgacode['Te'][None,:]*1e3  # keV --> eV

# set impurity species and sources rate
imp = namelist['imp'] = 'Ar'
namelist['source_type'] = 'const'
namelist['Phi0'] = 1e24

# Setup aurora sim to efficiently setup atomic rates and profiles over radius
asim = aurora.core.aurora_sim(namelist, geqdsk=geqdsk)

# plot normalized ionization frequency for the last time point, only within LCFS:
rhop_in = asim.rhop_grid[asim.rhop_grid<1.0]
S_z = asim.S_rates[asim.rhop_grid<1.0,:,-1] # take last time point
q_prof = interp1d(geqdsk['AuxQuantities']['RHOp'], geqdsk['QPSI'])(rhop_in)
Rhfs,Rlfs = aurora.grids_utils.get_HFS_LFS(geqdsk, rho_pol=rhop_in)
R_prof = (Rhfs+Rlfs)/2.   # take as average of flux surface

Ti_prof = asim.Te[-1, asim.rhop_grid<1.0] # use Ti=Te, only last time point

# inverse aspect ratio profile
eps_prof = (Rlfs-geqdsk['RMAXIS'])/geqdsk['RMAXIS'] # use LFS radius for considerations on trapped particles
nu_ioniz_star = aurora.atomic.plot_norm_ion_freq( S_z, q_prof, R_prof, asim.A_imp, Ti_prof,
                                                      rhop=rhop_in, plot=True, eps_prof=eps_prof)


# get average over charge states using fractional abundances in ionization equilibrium (no transport)
ne_avg = np.mean(kp['ne']['vals'],axis=0) # average over time
Te_avg = np.mean(kp['Te']['vals'],axis=0) # assume on the same radial basis as ne_avg

# get fractional abundances on ne (cm^-3) and Te (eV) grid
atom_data = aurora.get_atom_data(imp,['acd','scd'])
logTe, fz, rates = aurora.atomic.get_frac_abundances(atom_data, ne_avg, Te_avg, rho=rhop)

fz_profs = np.zeros_like(S_z)
for cs in np.arange(S_z.shape[1]):
    fz_profs[:,cs] = interp1d(rhop, fz[:,cs])(rhop_in)

nu_ioniz_star = aurora.atomic.plot_norm_ion_freq( S_z, q_prof, R_prof, asim.A_imp, Ti_prof,
                                                      nz_profs=fz_profs, rhop=rhop_in, plot=True,
                                                      eps_prof=eps_prof)
